from __future__ import annotations
from typing import Any, Callable, Iterable, Optional, Tuple, Union

import collections
import contextlib
import copy
import functools
import itertools
import math
import networkx as nx
import numpy as np
import os
import pathlib
import tempfile
import rdkit, rdkit.Chem, rdkit.Chem.AllChem
import scipy.interpolate, scipy.linalg, scipy.spatial


__all__ = [
    "closest_trio",
    "expand",
    "extrapolate",
    "interpolate",
    "IO",
    "lcm",
    "linearly_independent",
    "memoize",
    "mkdir_safe",
    "NestedDictionary",
    "nontrivial_vector",
    "normalize",
    "orthogonal_complement",
    "periodicity",
    "perpendicular_vector",
    "rotation_matrix",
    "rotation_matrix_m_to_n",
    "sample_spherical_triangle",
    "Settings",
    "spherical_excess",
    "temporary_seed",
    "work_dir",
    "xyz2mol",
]


def closest_trio(points):
    """
    Finds the three closest points from a given list of points.

    Parameters
    ----------
    points : numpy.ndarray
        3xNp Numpy array whose columns represent points on the unit sphere, where Np is the number of points.

    Returns
    -------
    numpy.ndarray:
        3x3 Numpy array whose columns are the three closest points from the given list of points.
    """
    tree = scipy.spatial.KDTree(points.T)
    pair = points.T[min(zip(*tree.query(points.T, k=2)), key=lambda t: t[0][1])[1]].T
    return points.T[
        min(zip(*tree.query(pair[:, 0, None].T, k=3)), key=lambda t: t[0][1])[1]
    ].T


def expand(path: str, dir: Optional[str] = None) -> str:
    p = pathlib.Path(path).expanduser()
    if dir is not None:
        p = pathlib.Path(dir).joinpath(p)
    return str(p.expanduser().resolve())


def extrapolate(x, y, x_extrap_to):
    # simple linear extrapolation
    x_lb, x_nlb, *_, x_nub, x_ub = x
    y_lb, y_nlb, *_, y_nub, y_ub = y

    # extrapolate down
    m_lb = (y_nlb - y_lb) / (x_nlb - x_lb)
    x_down = x_extrap_to[np.where(x_extrap_to < x_lb)]
    y_down = y_lb + m_lb * (x_down - x_lb)

    # extrapolate up
    m_ub = (y_ub - y_nub) / (x_ub - x_nub)
    x_up = x_extrap_to[np.where(x_extrap_to > x_ub)]
    y_up = y_ub + m_ub * (x_up - x_ub)

    return np.hstack((x_down, x, x_up)), np.hstack((y_down, y, y_up))


def interpolate(x, y, x_interp_to, method="cubic_spline"):
    if method == "sprague":
        return interpolate_sprague(x=x, y=y, x_interp_to=x_interp_to)
    elif method == "cubic_spline":
        return interpolate_cubic_spline(x=x, y=y, x_interp_to=x_interp_to)
    elif method == "linear_spline":
        return interpolate_linear_spline(x=x, y=y, x_interp_to=x_interp_to)
    else:
        raise ValueError(f"Interpolation method {method} not recognized.")


def interpolate_cubic_spline(x, y, x_interp_to):
    x_lb, *_, x_ub = x
    x_interp = x_interp_to[(x_interp_to >= x_lb) & (x_interp_to <= x_ub)]
    y_interp = scipy.interpolate.CubicSpline(x=x, y=y)(x_interp)

    return x_interp, y_interp


def interpolate_linear_spline(x, y, x_interp_to):
    x_lb, *_, x_ub = x
    x_interp = x_interp_to[(x_interp_to >= x_lb) & (x_interp_to <= x_ub)]
    y_interp = np.interp(x=x_interp, xp=x, fp=y)

    return x_interp, y_interp


def interpolate_sprague(x, y, x_interp_to):
    # FIXME: this implementation of Sprague interpolation is broken somehow!
    # data taken from https://link.springer.com/content/pdf/10.1007/978-3-642-27851-8_366-1.pdf
    # FIXME: we have to convert spec to wavelength spec with nanometer units first
    interp_coeffs = np.array(
        [
            [0, 0, 24, 0, 0, 0],
            [2, -16, 0, 16, -2, 0],
            [-1, -16, -30, 16, -1, 0],
            [-9, 39, 70, 66, -33, 7],
            [13, -64, 126, -124, 61, -12],
            [-5, 25, -50, 50, -25, 5],
        ]
    )
    # Sprague boundary points
    sbp1 = np.dot([884, -1960, 3033, -2648, 1080, 180], y[:6]) / 209
    sbp2 = np.dot([508, -540, 488, -367, 144, -24], y[:6]) / 209
    sbp3 = np.dot([-24, 144, -367, 488, -540, 508], y[-6:]) / 209
    sbp4 = np.dot([-180, 1080, -2648, 3033, -1960, 884], y[-6:]) / 209

    sprague_y = np.hstack([[sbp1, sbp2], y, [sbp3, sbp4]])[:, None]

    poly_c_mat = np.hstack(
        [
            (interp_coeffs @ np.roll(sprague_y, -i)[:6])
            for i in range(len(sprague_y) - 5)
        ]
    )  # [::-1]

    def interp(x_to):
        lower_bound_indices = np.searchsorted(x, x_to) - 1
        scaled_x = (x_to - x[lower_bound_indices]) / (
            x[lower_bound_indices + 1] - x[lower_bound_indices]
        )
        scaled_x_powers = np.vander(x=scaled_x, N=6)
        poly_coeffs = poly_c_mat[:, lower_bound_indices]
        # print(poly_coeffs.shape)
        # print(scaled_x.shape)
        return np.array(
            [
                np.polyval(p=coeffs[::-1], x=X)
                for coeffs, X in zip(poly_coeffs.T, scaled_x)
            ]
        )
        return (scaled_x_powers * poly_coeffs.T).sum(axis=1)

    y_interp = interp(x=x_interp_to)

    return x_interp_to, y_interp


_IO = collections.namedtuple("IOParams", ["work_dir", "inp", "out"])


class IO:
    def __init__(
        self,
        inp: Optional[str] = None,
        out: Optional[str] = None,
        work_dir: Optional[str] = ".",
        temp: Optional[bool] = False,
    ) -> None:
        self.inp = inp
        self.out = out
        self.work_dir = expand(work_dir)
        self.temp = temp

    @contextlib.contextmanager
    def __call__(self):
        mkdir_safe(self.work_dir)

        if self.temp:
            cm = tempfile.TemporaryDirectory(dir=self.work_dir)
        else:
            cm = contextlib.nullcontext(self.work_dir)

        with cm as wd:
            try:
                old_temp, self.temp = copy.copy(self.temp), False
                old_work_dir, self.work_dir = copy.copy(self.work_dir), wd

                yield _IO(
                    wd,
                    expand(self.inp, wd) if self.inp is not None else None,
                    expand(self.out, wd) if self.out is not None else None,
                )
            finally:
                self.temp, self.work_dir = old_temp, old_work_dir


def lcm(numbers: Iterable[int]) -> int:
    a, *b = numbers
    if len(b) > 1:
        return lcm(numbers=(a, lcm(numbers=b)))
    else:
        [b] = b
        return a * b // math.gcd(a, b)


def linearly_independent(vectors, indep=None):
    # vectors is kxNv array where Nv is number of vectors
    # indep is kxNi array where Ni is number of linearly independent vectors

    if indep is None:
        indep = np.array([])

    k, *_ = vectors.shape  # dimension of vectors
    *_, n = indep.shape  # number of independent vectors

    if n == k:
        return indep
    else:
        array_generator = (np.vstack([*indep.T, v]) for v in vectors.T)
        try:
            indep = next(
                a
                for a in array_generator
                if scipy.linalg.null_space(a).shape[-1] == k - n - 1
            ).T
            return linearly_independent(vectors=vectors, indep=indep)
        except StopIteration:
            return indep


class _Cache(collections.OrderedDict):
    def last_result(self, n=1):
        return tuple(self.values())[-n]

    def last_args(self, n=1):
        args, kwarg_items = tuple(self.keys())[-n]
        return args, dict(kwarg_items)


def memoize(func):
    func.cache = _Cache()

    @functools.wraps(func)
    def memoized(*args, **kwargs):
        k = (args, frozenset(kwargs.items()))
        if k not in func.cache:
            func.cache[k] = func(*args, **kwargs)

        return func.cache[k]

    return memoized


def mkdir_safe(dir: str) -> None:
    with contextlib.suppress(FileExistsError):
        os.makedirs(dir)


def nontrivial_vector(R, seed: Optional[int] = None):
    """
    Generates a vector which is acted upon nontrivially (i.e. is sent to a linearly independent vector) by the given rotation matrix.

    Parameters
    ----------
    R: numpy.ndarray
        3x3 numpy array representing a rotation matrix.

    Returns
    -------
    numpy.ndarray:
        3x1 numpy array representing a vector which is acted upon nontrivially by R.
    """
    if (
        np.allclose(R, np.eye(R.shape[0]))
        or np.allclose(R, -np.eye(R.shape[0]))
        or np.allclose(R, np.zeros(R.shape[0]))
    ):
        return None

    # get the eigenvectors with real (i.e. 1 or -1) eigenvalues, since these are mapped to colinear vectors by R
    # each column of evecs is an eigenvector of R
    evals, evecs = scipy.linalg.eig(R)

    # FIXME: why np.real(evec) instead of simply evec?
    real_eigenbasis = np.real(evecs.T[np.isclose(np.imag(evals), 0)].T)

    # form the linear combination of the "trivial" eigenvectors
    # get random coefficients between 1 and 2 so that 0 is never chosen
    # the result is guaranteed to be mapped to a linearly independent vector
    # by R because the "trivial" eigenvectors do not all have the same eigenvalues #all have different eigenvalues
    # this is true because R is not proportional to the identity matrix
    with temporary_seed(seed=seed):
        coeffs = np.random.uniform(low=1, high=2, size=(real_eigenbasis.shape[1], 1))

    return normalize(v=real_eigenbasis @ coeffs)


def normalize(v):
    norm = scipy.linalg.norm(v)

    return v / norm if norm > 0 else v


def orthogonal_complement(v, l):
    return normalize(v - np.dot(v.T, l) * l)


def periodicity(matrix) -> int:
    # if A is periodic, then its eigenvalues are roots of unity, and its periodicity is the lcm of the periodicities of these roots of unity
    # kth roots of unity form the vertices of a regular k-gon with internal angles 2*pi/k
    # the angle between two such vertices z1=a+jb and z2=c+jd is given by cos(theta) = a*c + b*d = Re(z1*conj(z2))
    # choose z2 = z1**2 (clearly z2 is still a root of unity); then z1*conj(z2) = exp(2*pi*j/k)*exp(-4*pi*j/k) = exp(-2*pi*j/k)
    # then Re(z1*conj(z2)) = Re(exp(-2*pi*j/k)) = cos(2*pi*j/k) = Re(z1)
    # so 2*pi*j/k = arccos(Re(z1)) -> j/k = arccos(Re(z1))/(2*pi), and k = lcm(k/j1, k/j2,...)
    evals = scipy.linalg.eigvals(matrix)
    angles = (max(min(z.real, 1), -1) for z in evals if not np.isclose(z, 1))
    # if z is close to 1, then it contributes a period of 1, which doesn't impact the lcm and therefore the final period
    periods = (int((2 * np.pi / np.arccos(angle)).round()) for angle in angles)

    try:
        return lcm(numbers=periods)
    except ValueError:
        # `periods` must not have any values in it, so all evals must have been close to 1
        return 1


def perpendicular_vector(a, b=None):
    """
    Generates a unit vector which is perpendicular to one or two given nonzero vector(s).

    Parameters
    ----------
    a: numpy.ndarray
        3x1 Numpy array representing a nonzero vector.
    b: numpy.ndarray
        3x1 Numpy array representing a nonzero vector.

    Returns
    -------
    numpy.ndarray:
        3x1 Numpy array representing a unit vector which is perpendicular to a (and b, if applicable).
    """
    if b is None:
        m = np.zeros(a.shape)

        # storing in variable for reuse
        ravel_a = np.ravel(a)

        # index of the first nonzero element of a
        i = (ravel_a != 0).argmax()
        # first index of a which is not i
        j = next(ind for ind in range(len(ravel_a)) if ind != i)
        # unravel indices for 3x1 arrays m and a
        i, j = (
            np.unravel_index(i, a.shape),
            np.unravel_index(j, a.shape),
        )

        # make m = np.array([[-ay,ax,0]]).T so np.dot(m.T,a) = -ax*ay + ax*ay = 0
        m[j] = a[i]
        m[i] = -a[j]
    else:
        m = np.cross(a.T, b.T).T

    return normalize(v=m)


def rotation_matrix(axis, cos_theta):
    """
    Generates a matrix which rotates a vector a given angle about a given axis.

    Parameters
    ----------
    axis: numpy.ndarray
        3x1 Numpy array representing the vector whose direction is the axis of rotation.
    cos_theta: float
        Cosine of angle of rotation about the axis of rotation.

    Returns
    -------
    numpy.ndarray:
        3x3 Numpy array representing the rotation matrix which rotates a vector by the given angle about the given axis.
    """
    u1, u2, u3 = np.ravel(normalize(v=axis))

    K = np.array([[0, -u3, u2], [u3, 0, -u1], [-u2, u1, 0]])

    sin_theta = np.sqrt(1 - cos_theta ** 2)

    return (np.eye(3) + sin_theta * K + (1 - cos_theta) * (K @ K)).astype("float64")


def rotation_matrix_m_to_n(m, n):
    """
    Generates a matrix which rotates one given vector to another.

    Parameters
    ----------
    m: numpy.ndarray
        3x1 Numpy array representing the vector to be rotated.
    n: numpy.ndarray
        3x1 Numpy array representing the vector after rotation, i.e. the target vector.

    Returns
    -------
    numpy.ndarray:
        3x3 Numpy array representing the rotation matrix which maps m to n.
    """
    # make rotation axis, which is perpendicular to both m and n
    # no need to normalize, since rotation_matrix will do that for us
    u = np.cross(m.T, n.T).T

    # cosine of the angle between m and n
    c = np.dot(normalize(v=m).T, normalize(v=n))

    return rotation_matrix(axis=u, cos_theta=c)


def sample_spherical_triangle(A, B, C, sin_alpha, sin_beta, sin_gamma, seed=None):
    # see https://www.graphics.cornell.edu/pubs/1995/Arv95c.pdf
    # a, b, and c are cross products of normal vectors, so their magnitudes
    # are the sine of the angles between these normal vectors; these angles
    # are also the angles between planes and therefore the great arcs which
    # define the legs of the triangle; therefore, these angles are also
    # the internal angles of the triangle
    eps = np.finfo(float).eps  # machine precision

    with temporary_seed(seed=seed):
        fraction, cos_theta = np.random.uniform(low=eps, high=1, size=2)

    cos_alpha, cos_beta, cos_gamma = np.sqrt(
        (1 - sin_alpha ** 2, 1 - sin_beta ** 2, 1 - sin_gamma ** 2)
    )

    area = fraction * spherical_excess(
        cos_alpha=cos_alpha, cos_beta=cos_beta, cos_gamma=cos_gamma
    )
    cos_area, sin_area = np.cos(area), np.sin(area)

    s = sin_area * cos_alpha - cos_area * sin_alpha  # sin(area - alpha)
    t = cos_area * cos_alpha + sin_area * sin_alpha  # cos(area - alpha)
    u = t - cos_alpha
    v = s + (cos_gamma + cos_beta * cos_alpha) / sin_beta  # spherical law of cosines

    q = ((v * t - u * s) * cos_alpha - v) / ((v * s + u * t) * sin_alpha)
    C_prime = q * A + np.sqrt(1 - q ** 2) * orthogonal_complement(v=C, l=A)

    z = 1 - cos_theta * (1 - np.dot(C_prime.T, B))
    return z * B + np.sqrt(1 - z ** 2) * orthogonal_complement(v=C_prime, l=B)


def spherical_excess(cos_alpha: float, cos_beta: float, cos_gamma: float) -> float:
    # Girard's formula for spherical excess
    return np.arccos((cos_alpha, cos_beta, cos_gamma)).sum() - np.pi


@contextlib.contextmanager
def temporary_seed(seed: int) -> None:
    # adapted from https://stackoverflow.com/a/49557127
    state = np.random.get_state()
    if seed is not None:
        np.random.seed(seed)
    try:
        yield
    finally:
        np.random.set_state(state)


@contextlib.contextmanager
def work_dir(dir: Optional[str] = None):
    # FIXME: learn how to type annotation yields
    with contextlib.nullcontext(
        expand(dir)
    ) if dir is not None else tempfile.TemporaryDirectory(dir=dir) as wd:
        try:
            mkdir_safe(dir=wd)
            yield wd
        finally:
            pass


class NestedDictionary(collections.abc.MutableMapping):
    def __init__(self, *args, **kwargs) -> None:
        self.d = dict(*args, **kwargs)

    def __getitem__(self, keys) -> NestedDictionary:
        if not isinstance(keys, tuple):
            keys = (keys,)

        branch = self.d
        for k in keys:
            branch = branch[k]

        return NestedDictionary(branch) if isinstance(branch, dict) else branch

    def __setitem__(self, keys, value) -> None:
        if not isinstance(keys, tuple):
            keys = (keys,)

        *most_keys, last_key = keys

        branch = self.d
        for k in most_keys:
            if k not in branch:
                branch[k] = {}
            branch = branch[k]

        branch[last_key] = value

    def __delitem__(self, keys) -> None:
        if not isinstance(keys, tuple):
            keys = (keys,)

        *most_keys, last_key = keys

        branch = self.d
        for k in most_keys:
            # FIXME: does this really need to be here? if we're deleting the item, we shouldn't ever need to make branches on the way to the item...
            if k not in branch:
                branch[k] = {}
            branch = branch[k]

        del branch[last_key]

    def __iter__(self, d=None, prepath=()):
        if d == None:
            d = self.d
        for k, v in d.items():
            if hasattr(v, "items"):
                for keys in self.__iter__(d=v, prepath=prepath + (k,)):
                    yield keys
            else:
                yield prepath + (k,)

    def __len__(self) -> int:
        return sum(1 for _ in self)

    def __str__(self) -> str:
        return str(self.d)

    def __repr__(self) -> str:
        return repr(self.d)


class Settings(NestedDictionary):
    def update(self, settings: Settings) -> None:
        for k, v in settings.items():
            self.__setitem__(keys=k, value=v)


# ------------------------------------------------------ #

##
# Written by Jan H. Jensen based on this paper Yeonjoon Kim and Woo Youn Kim
# "Universal Structure Conversion Method for Organic Molecules: From Atomic Connectivity
# to Three-Dimensional Geometry" Bull. Korean Chem. Soc. 2015, Vol. 36, 1769-1777 DOI: 10.1002/bkcs.10334
#


def xyz2mol(
    atomic_numbers, charge, atomic_positions, charged_fragments: Optional[bool] = True
) -> rdkit.Chem.rdchem.Mol:
    # alternatively radicals are made if charged_fragments=False

    # Get atom connectivity (ac) matrix, list of atomic numbers, molecular charge,
    # and mol object with no connectivity information
    ac, mol = _atom_connectivity(atomic_numbers, atomic_positions)

    # Convert ac to bond order matrix and add connectivity and charge info to mol object
    new_mol = _connected_mol(mol, ac, atomic_numbers, charge, charged_fragments)

    # Check for stereocenters and chiral centers
    new_mol = _chiral_stereo_check(new_mol)

    # # Canonical hack
    # smiles = Chem.MolToSmiles(mol, isomericSmiles=True)
    # m = Chem.MolFromSmiles(smiles)
    # smiles = Chem.MolToSmiles(m, isomericSmiles=True)

    return new_mol


def _atom_connectivity(atomic_numbers: Iterable[int], atomic_positions):
    # generate proto molecule
    first, *rest = atomic_numbers

    mol = rdkit.Chem.MolFromSmarts("[#" + str(first) + "]")
    rwMol = rdkit.Chem.RWMol(mol)

    for Z in rest:
        rwMol.AddAtom(rdkit.Chem.Atom(Z))

    mol = rwMol.GetMol()

    # compute atom connectivity matrix

    dmat = scipy.spatial.distance_matrix(atomic_positions, atomic_positions)
    pt = rdkit.Chem.GetPeriodicTable()

    num_atoms = len(atomic_numbers)
    ac = np.zeros((num_atoms, num_atoms)).astype(int)

    for i in range(num_atoms):
        a_i = mol.GetAtomWithIdx(i)
        Rcov_i = pt.GetRcovalent(a_i.GetAtomicNum()) * 1.30
        for j in range(i + 1, num_atoms):
            a_j = mol.GetAtomWithIdx(j)
            Rcov_j = pt.GetRcovalent(a_j.GetAtomicNum()) * 1.30
            if dmat[i, j] <= Rcov_i + Rcov_j:
                ac[i, j] = 1
                ac[j, i] = 1

    return ac, mol


def _connected_mol(
    mol: rdkit.Chem.rdchem.Mol,
    ac,
    atomic_numbers: Iterable[int],
    charge: int,
    charged_fragments: bool,
) -> rdkit.Chem.rdchem.Mol:
    # convert ac matrix to bond order (BO) matrix
    BO, atomic_valence_electrons = _bond_order(
        ac, atomic_numbers, charge, charged_fragments
    )

    # add BO connectivity and charge info to mol object
    mol = _bond_order_to_mol(
        mol, BO, atomic_numbers, atomic_valence_electrons, charge, charged_fragments
    )

    return mol


def _bond_order(
    ac, atomic_numbers: Iterable[int], charge: int, charged_fragments: bool
):
    # TODO
    atomic_valence = collections.defaultdict(list)
    atomic_valence[1] = [1]
    atomic_valence[6] = [4]
    atomic_valence[7] = [4, 3]
    atomic_valence[8] = [2, 1]
    atomic_valence[9] = [1]
    atomic_valence[14] = [4]
    atomic_valence[15] = [5, 4, 3]
    atomic_valence[16] = [6, 4, 2]
    atomic_valence[17] = [1]
    atomic_valence[32] = [4]
    atomic_valence[35] = [1]
    atomic_valence[53] = [1]

    atomic_valence_electrons = {}
    atomic_valence_electrons[1] = 1
    atomic_valence_electrons[6] = 4
    atomic_valence_electrons[7] = 5
    atomic_valence_electrons[8] = 6
    atomic_valence_electrons[9] = 7
    atomic_valence_electrons[14] = 4
    atomic_valence_electrons[15] = 5
    atomic_valence_electrons[16] = 6
    atomic_valence_electrons[17] = 7
    atomic_valence_electrons[32] = 4
    atomic_valence_electrons[35] = 7
    atomic_valence_electrons[53] = 7

    # make a list of valences, e.g. for CO: [[4],[2,1]]
    valences_list_of_lists = [atomic_valence[Z] for Z in atomic_numbers]

    # convert [[4],[2,1]] to [[4,2],[4,1]]
    valences_list = itertools.product(*valences_list_of_lists)

    best_BO = ac.copy()

    # implemenation of algorithm shown in Figure 2
    # ua: unsaturated atoms
    # du: degree of unsaturation (u matrix in Figure)
    # best_BO: Bcurr in Figure
    #

    for valences in valences_list:
        ac_valence = list(ac.sum(axis=1))
        ua, du_from_ac = _get_unsaturated_atoms(valences, ac_valence)

        if len(ua) == 0 and _bond_order_is_ok(
            ac,
            ac,
            charge,
            du_from_ac,
            atomic_valence_electrons,
            atomic_numbers,
            charged_fragments,
        ):
            return ac, atomic_valence_electrons

        for ua_pairs in _get_ua_pairs(ua, ac):
            BO = _get_bond_orders(ac, ua, du_from_ac, valences, ua_pairs)
            if _bond_order_is_ok(
                BO,
                ac,
                charge,
                du_from_ac,
                atomic_valence_electrons,
                atomic_numbers,
                charged_fragments,
            ):
                return BO, atomic_valence_electrons

            elif BO.sum() >= best_BO.sum() and (BO.sum(axis=1) <= valences).all():
                best_BO = BO.copy()

    return best_BO, atomic_valence_electrons


def _get_unsaturated_atoms(max_valences, valences):
    diffs = np.array(max_valences) - np.array(valences)
    inds = diffs > 0

    unsatured_atoms = np.flatnonzero(inds)
    degree_of_unsaturation = diffs[inds]

    return unsatured_atoms, degree_of_unsaturation


def _bond_order_is_ok(
    BO,
    ac,
    charge: int,
    du,
    atomic_valence_electrons,
    atomic_numbers: Iterable[int],
    charged_fragments: bool,
) -> bool:
    Q = 0  # total charge
    q_list = []
    if charged_fragments:
        BO_valences = list(BO.sum(axis=1))
        for i, atom in enumerate(atomic_numbers):
            q = _get_atomic_charge(atom, atomic_valence_electrons[atom], BO_valences[i])
            Q += q
            if atom == 6:
                number_of_single_bonds_to_C = list(BO[i, :]).count(1)
                if number_of_single_bonds_to_C == 2 and BO_valences[i] == 2:
                    Q += 1
                    q = 2
                if number_of_single_bonds_to_C == 3 and Q + 1 < charge:
                    Q += 2
                    q = 1

            if q != 0:
                q_list.append(q)

    return (BO - ac).sum() == sum(du) and charge == Q and len(q_list) <= abs(charge)


def _get_bond_orders(ac, ua, du, valences, ua_pairs):
    BO = ac.copy()
    du_save = []

    while du_save != du:
        for i, j in ua_pairs:
            BO[i, j] += 1
            BO[j, i] += 1

        BO_valence = list(BO.sum(axis=1))
        du_save = copy.copy(du)
        ua, du = _get_unsaturated_atoms(valences, BO_valence)
        ua_pairs, *_ = _get_ua_pairs(ua, ac)

    return BO


def _chiral_stereo_check(mol: rdkit.Chem.rdchem.Mol) -> rdkit.Chem.rdchem.Mol:
    rdkit.Chem.SanitizeMol(mol)
    rdkit.Chem.DetectBondStereochemistry(mol, -1)
    rdkit.Chem.AssignStereochemistry(mol, flagPossibleStereoCenters=True, force=True)
    rdkit.Chem.AssignAtomChiralTagsFromStructure(mol, -1)

    return mol


def _get_atomic_charge(atom, atomic_valence_electrons, BO_valence) -> int:
    if atom == 1:
        charge = 1 - BO_valence
    elif atom == 5:
        charge = 3 - BO_valence
    elif atom == 15 and BO_valence == 5:
        charge = 0
    elif atom == 16 and BO_valence == 6:
        charge = 0
    else:
        charge = atomic_valence_electrons - 8 + BO_valence

    return charge


def _clean_charges(mol: rdkit.Chem.rdchem.Mol) -> rdkit.Chem.rdchem.Mol:
    rdkit.Chem.SanitizeMol(mol)
    # rxn_smarts = ['[N+:1]=[*:2]-[C-:3]>>[N+0:1]-[*:2]=[C-0:3]',
    #              '[N+:1]=[*:2]-[O-:3]>>[N+0:1]-[*:2]=[O-0:3]',
    #              '[N+:1]=[*:2]-[*:3]=[*:4]-[O-:5]>>[N+0:1]-[*:2]=[*:3]-[*:4]=[O-0:5]',
    #              '[#8:1]=[#6:2]([!-:6])[*:3]=[*:4][#6-:5]>>[*-:1][*:2]([*:6])=[*:3][*:4]=[*+0:5]',
    #              '[O:1]=[c:2][c-:3]>>[*-:1][*:2][*+0:3]',
    #              '[O:1]=[C:2][C-:3]>>[*-:1][*:2]=[*+0:3]']

    rxn_smarts = [
        "[#6,#7:1]1=[#6,#7:2][#6,#7:3]=[#6,#7:4][CX3-,NX3-:5][#6,#7:6]1=[#6,#7:7]>>\
                   [#6,#7:1]1=[#6,#7:2][#6,#7:3]=[#6,#7:4][-0,-0:5]=[#6,#7:6]1[#6-,#7-:7]",
        "[#6,#7:1]1=[#6,#7:2][#6,#7:3](=[#6,#7:4])[#6,#7:5]=[#6,#7:6][CX3-,NX3-:7]1>>\
                   [#6,#7:1]1=[#6,#7:2][#6,#7:3]([#6-,#7-:4])=[#6,#7:5][#6,#7:6]=[-0,-0:7]1",
    ]

    fragments = rdkit.Chem.GetMolFrags(mol, asMols=True, sanitizeFrags=False)

    for i, fragment in enumerate(fragments):
        for smarts in rxn_smarts:
            patt = rdkit.Chem.MolFromSmarts(smarts.split(">>")[0])
            while fragment.HasSubstructMatch(patt):
                rxn = AllChem.ReactionFromSmarts(smarts)
                ps = rxn.RunReactants((fragment,))
                fragment = ps[0][0]
                rdkit.Chem.SanitizeMol(fragment)

        mol = fragment if i == 0 else rdkit.Chem.CombineMols(mol, fragment)

    return mol


def _bond_order_to_mol(
    mol: rdkit.Chem.rdchem.Mol,
    BO_matrix,
    atomic_numbers: Iterable[int],
    atomic_valence_electrons,
    mol_charge,
    charged_fragments: bool,
):
    # based on code written by Paolo Toscani

    l = len(BO_matrix)
    l2 = len(atomic_numbers)
    BO_valences = list(BO_matrix.sum(axis=1))

    if l != l2:
        raise RuntimeError(
            "sizes of adjMat ({0:d}) and atomic_numbers " "{1:d} differ".format(l, l2)
        )

    rwMol = rdkit.Chem.RWMol(mol)

    bondTypeDict = {
        1: rdkit.Chem.BondType.SINGLE,
        2: rdkit.Chem.BondType.DOUBLE,
        3: rdkit.Chem.BondType.TRIPLE,
    }

    for i in range(l):
        for j in range(i + 1, l):
            bo = int(round(BO_matrix[i, j]))
            if bo == 0:
                continue
            bt = bondTypeDict.get(bo, rdkit.Chem.BondType.SINGLE)
            rwMol.AddBond(i, j, bt)
    mol = rwMol.GetMol()

    if charged_fragments:
        mol = set_atomic_charges(
            mol,
            atomic_numbers,
            atomic_valence_electrons,
            BO_valences,
            BO_matrix,
            mol_charge,
        )
    else:
        mol = set_atomic_radicals(
            mol, atomic_numbers, atomic_valence_electrons, BO_valences
        )

    return mol


def set_atomic_charges(
    mol: rdkit.Chem.rdchem.Mol,
    atomic_numbers: Iterable[int],
    atomic_valence_electrons,
    BO_valences,
    BO_matrix,
    mol_charge,
) -> rdkit.Chem.rdchem.Mol:
    q = 0
    for i, atom in enumerate(atomic_numbers):
        a = mol.GetAtomWithIdx(i)
        charge = _get_atomic_charge(
            atom, atomic_valence_electrons[atom], BO_valences[i]
        )
        q += charge
        if atom == 6:
            number_of_single_bonds_to_C = list(BO_matrix[i, :]).count(1)
            if number_of_single_bonds_to_C == 2 and BO_valences[i] == 2:
                q += 1
                charge = 0
            if number_of_single_bonds_to_C == 3 and q + 1 < mol_charge:
                q += 2
                charge = 1

        if abs(charge) > 0:
            a.SetFormalCharge(int(charge))

    mol = _clean_charges(mol)

    return mol


def set_atomic_radicals(
    mol: rdkit.Chem.rdchem.Mol,
    atomic_numbers: Iterable[int],
    atomic_valence_electrons,
    BO_valences,
) -> rdkit.Chem.rdchem.Mol:
    # The number of radical electrons = absolute atomic charge
    for i, atom in enumerate(atomic_numbers):
        a = mol.GetAtomWithIdx(i)
        charge = _get_atomic_charge(
            atom, atomic_valence_electrons[atom], BO_valences[i]
        )

        if abs(charge) > 0:
            a.SetNumRadicalElectrons(abs(int(charge)))

    return mol


def _get_ua_pairs(ua, ac):
    bonds = []

    for k, i in enumerate(ua):
        for j in ua[k + 1 :]:
            if ac[i, j] == 1:
                bonds.append(tuple(sorted([i, j])))

    if len(bonds) == 0:
        return [()]

    G = nx.Graph()
    G.add_edges_from(bonds)
    ua_pairs = [list(nx.max_weight_matching(G))]

    return ua_pairs


# ----------------------------------------------------- #

# def plot_lune(self, samples):
#     import matplotlib.pyplot as plt
#     from mpl_toolkits.mplot3d import Axes3D
#     xs,ys,zs = samples
#     fig = plt.figure()
#     ax = Axes3D(fig)
#     phigrid,thetagrid = np.mgrid[0.0:np.pi:100j,0.0:2.0*np.pi:100j]
#     xgrid = np.sin(phigrid)*np.cos(thetagrid)
#     ygrid = np.sin(phigrid)*np.sin(thetagrid)
#     zgrid = np.cos(phigrid)
#     ax.plot_surface(xgrid,ygrid,zgrid,rstride=1,cstride=1,color='c',alpha=0.1,linewidth=0)
#     ax.scatter(xs,ys,zs)
#     ax.quiver([0,0,0,0,0,0],[0,0,0,0,0,0],[0,0,0,0,0,0],[1,0,0,-1,0,0],[0,1,0,0,-1,0],[0,0,1,0,0,-1],arrow_length_ratio=0.1,linewidths=[2,2,2,2,2,2])
#     #ax.set_aspect('equal')
#     ax.xaxis._axinfo['juggled'] = (0,0,0)
#     ax.yaxis._axinfo['juggled'] = (1,1,1)
#     ax.zaxis._axinfo['juggled'] = (2,2,2)
#     plt.show(block=True)



# import dlib
# import numpy as np
# import scipy.interpolate
import re

#
import materia

#
from materia.handler import Handler
from materia.actions import (
    QChemModifyRSHParameter,
    QChemIncreaseResponseIterations,
    QChemIncreaseSCFIterations,
    Rerun,
)

#
__all__ = [
    "MinimizeKoopmanError",
    "QChemResponseDIISConvergence",
    "QChemSCFConvergence",
]

# import materia, numpy as np, scipy
#
# def minimize_m_n(m, n, x_min, x_max, num_xs, round=3):
#     xs = np.linspace(x_min,x_max,num_xs)
#     ms = [Power(x=x,p=m) for x in xs]
#     ns = [Power(x=x,p=n) for x in xs]
#     subs = [Subtract() for x in xs]
#     for a,b,sub in zip(ms,ns,subs):
#         sub.requires(a=a,b=b)
#     min = Minimum()
#     min.requires(**{f'f{i}': sub for i,sub in enumerate(subs)})
#     f0,x0 = materia.Workflow(*ms,*ns,*subs,min).run()[3*num_xs]
#     return x0,f0,np.round(x0,round) == np.power(n/m,1/(m-n)).round(round)
#
# class Power(materia.Task):
#     def __init__(self, x, p, handlers=None, name=None):
#         super().__init__(handlers=handlers,name=name)
#         self.x = x
#         self.p = p
#     def run(self):
#         return self.x, self.x**self.p
#
# class Subtract(materia.Task):
#     def run(self, a, b):
#         x,a = a
#         x,b = b
#         return x, a - b
#
# class Minimum(materia.Task):
#     def run(self, **kwargs):
#         xs,fs = zip(*kwargs.values())
#         spline = scipy.interpolate.CubicSpline(xs,fs)
#         candidates = spline.derivative().roots()
#         return min(zip(spline(candidates),candidates))


class MinimizeKoopmanError(Handler):
    def __init__(
        self, omega_min, omega_max, squared_error_threshold=None, input_ids=None
    ):  # FIXME: squared_error -> 1e-9?
        self.omega_min = omega_min
        self.omega_max = omega_max
        self.squared_error_threshold = (
            squared_error_threshold
            if squared_error_threshold is not None
            else materia.Qty(value=1e-3, unit=materia.eV)
        )
        self.koopman_errors = {}

        self.input_ids = (
            tuple(input_ids) if input_ids is not None else ("omega", "koopman_error")
        )

    def check(self, result, task):
        omega, koopman_error = (task_outputs[k] for k in self.input_ids)
        self.koopman_errors[omega] = koopman_error
        return koopman_error > self.squared_error_threshold

    def _get_next_omega(self):
        def _proxy(x):
            try:
                return self.koopman_errors[x].value
            except KeyError:
                raise ValueError(f"{x}")

        try:
            dlib.find_min_global(
                _proxy,
                [self.omega_min,],
                [self.omega_max,],
                len(self.koopman_errors) + 1,
            )
        except ValueError as e:
            return float(str(e))

    def handle(self, result, task):
        raise NotImplementedError
        # create five tasks: omega, gs, cation, anion, koopman_error
        ke = ComputeKoopmanError()
        # links tasks: gs.requires(omega=omega),cation.requires(omega=omega),anion.requires(omega=omega),koopman_error.requires(gs=gs,cation=cation,anion=anion)
        return [InsertTasks(tasks=(omega, gs, cation, anion, koopman_error))]
        # return [QChemModifyRSHParameter(omega=self._get_next_omega()),Rerun()]
        # error_curve_spline = scipy.interpolate.CubicSpline
        # def find_min(f,bounds):
        # x_min,x_max = bounds
        # y = np.linspace(x_min,x_max,5)
        # [x0,],_ = dlib.find_min_global(lambda a: scipy.interpolate.CubicSpline(y,f(y))(a),[min(y),],[max(y),],100)
        # delta = (x_max - x_min)/100
        # z = np.linspace(x0-delta,x0+delta,5)
        # return np.hstack([y,z]),np.hstack([f(y),f(z)]),dlib.find_min_global(lambda a: scipy.interpolate.CubicSpline(z,f(z))(a),[min(z),],[max(z),],100)[0][0]#scipy.optimize.minimize_scalar(scipy.interpolate.CubicSpline(z,f(z)),(min(z),max(z)),method='golden')['x']


class QChemResponseDIISConvergence(Handler):
    def __init__(self, increase_factor=2):
        self.increase_factor = increase_factor

    def check(self, result, task):
        with open(task.io.out, "r") as f:
            if re.search(
                r"DIIS\s*failed\s*to\s*converge\s*within\s*the\s*given\s*number\s*of\s*iterations",
                "".join(f.readlines()),
            ):
                return True

        return False

    def handle(self, result, task):
        return [
            QChemIncreaseResponseIterations(increase_factor=self.increase_factor),
            Rerun(),
        ]


class QChemSCFConvergence(Handler):
    def __init__(self, increase_factor=2):
        self.increase_factor = increase_factor

    def check(self, result, task):
        with open(task.output_path, "r") as f:
            if re.search(
                r"gen_scfman_exception:\s*SCF\s*failed\s*to\s*converge",
                "".join(f.readlines()),
            ):
                return True

        return False

    def handle(self, result, task):
        return [
            QChemIncreaseSCFIterations(increase_factor=self.increase_factor),
            Rerun(),
        ]



# from __future__ import annotations
# import os
# import subprocess

# from ...workflow.tasks.task import Task

# __all__ = ["ExecuteNWChem"]


# class ExecuteNWChem(Task):
#     def __init__(
#         self,
#         input_name,
#         output_name,
#         executable="nwchem",
#         work_directory=".",
#         num_cores=1,
#         parallel=False,
#         handlers=None,
#         name=None,
#     ):
#         super().__init__(handlers=handlers)
#         self.settings["input_path"] = materia.expand(
#             os.path.join(work_directory, input_name)
#         )
#         self.settings["output_path"] = materia.expand(
#             os.path.join(work_directory, output_name)
#         )
#         self.settings["scratch_directory"] = materia.expand(scratch_directory)

#         self.settings["executable"] = executable
#         self.settings["work_directory"] = materia.expand(work_directory)

#         self.settings["num_cores"] = num_cores
#         self.settings["parallel"] = parallel

#         try:
#             os.makedirs(self.settings["work_directory"])
#         except FileExistsError:
#             pass

#     def run(self):
#         with open(self.settings["output_path"], "w") as f:
#             if self.settings["parallel"]:
#                 subprocess.call(
#                     [
#                         "mpirun",
#                         "-np",
#                         str(self.settings["num_cores"]),
#                         self.settings["executable"],
#                         self.settnigs["input_path"],
#                     ],
#                     stdout=f,
#                 )
#             else:
#                 subprocess.call(
#                     [self.settings["executable"], self.settings["input_path"]], stdout=f
#                 )


# # from ..calculation import Calculation
# #
# # class NWChemPolarizability(Calculation):
# #     def __init__(self):
# #         raise NotImplementedError
# #
# # class NWChemRTTDDFT(Calculation):
# #     def __init__(self):
# #         raise NotImplementedError
# #
# # def compute_polarizability(self, molecule, num_cores, parallel, header_params={},\
# #                            geometry_params={}, set_geometry_params={}, basis_params={},\
# #                            dft_params={}, property_params={}, task_params={}):
# #
# #     input_path = os.path.join(self.scratch_directory,'polarizability.in')
# #     output_path = os.path.join(self.scratch_directory,'polarizability.out')
# #
# #     default_header_params = {'title':'molecule','echo':True,'scratch_dir':'./scratch','permanent_dir':'./permanent','start': True}
# #     merged_header_params = self._merge_new_params(old_params=default_header_params,new_params=header_params)
# #
# #     default_geometry_params = {'title': 'system', 'xyz_filepath': molecule.xyz.filepath}
# #     merged_geometry_params = self._merge_new_params(old_params=default_geometry_params,new_params=geometry_params)
# #
# #     default_set_geometry_params = {'set_params': {'variable_name': 'geometry', 'variable_value': '\"system\"'}}
# #     merged_set_geometry_params = self._merge_new_params(old_params=default_set_geometry_params,new_params=set_geometry_params)
# #
# #     default_basis_params = {'atom_strings': ('*',), 'library_strings': ('6-311G**',)}
# #     merged_basis_params = self._merge_new_params(old_params=default_basis_params,new_params=basis_params)
# #
# #     default_dft_params = {'direct': True, 'xc_params': {'xc_functional_strings': ('b3lyp',)}}
# #     merged_dft_params = self._merge_new_params(old_params=default_dft_params,new_params=dft_params)
# #
# #     default_property_params = {'response_order': 1, 'response_frequency': 0}
# #     merged_property_params = self._merge_new_params(old_params=default_property_params,new_params=property_params)
# #
# #     default_task_params = {'task_params': {'theory_level': 'dft', 'operation': 'property'}}
# #     merged_task_params = self._merge_new_params(old_params=default_task_params,new_params=task_params)
# #
# #     input = NWChemInput()
# #     input.add_block(top=merged_header_params)
# #     input.add_block(geometry=merged_geometry_params)
# #     input.add_block(top=merged_set_geometry_params)
# #     input.add_block(basis=merged_basis_params)
# #     input.add_block(dft=merged_dft_params)
# #     input.add_block(property=merged_property_params)
# #     input.add_block(top=merged_task_params)
# #     input.write_input(filename=input_path)
# #
# #     self.execute(input_path=input_path,output_path=output_path,num_cores=num_cores,parallel=parallel)
# #
# #     output = NWChemOutput(filename=output_path)
# #
# #     molecule.polarizability = output.get_polarizability()
# #
# # def compute_dynamic_dipole_moment(self, molecule, num_cores, parallel,
# #                                   x=True, y=True, z=True, header_params={},
# #                                   geometry_params={}, set_geometry_params={},
# #                                   basis_params={}, dft_params={}, dft_task_params={},
# #                                   rttddft_x_params={}, rttddft_x_task_params={},
# #                                   rttddft_y_params={}, rttddft_y_task_params={},
# #                                   rttddft_z_params={}, rttddft_z_task_params={}):
# #
# #     input_path = os.path.join(self.scratch_directory,'dynamic_dipole_moment.in')
# #     output_path = os.path.join(self.scratch_directory,'dynamic_dipole_moment.out')
# #
# #     input = NWChemInput()
# #
# #     default_header_params = {'title':'molecule','echo':True,'scratch_dir':'./scratch','permanent_dir':'./permanent','start': True}
# #     merged_header_params = self._merge_new_params(old_params=default_header_params,new_params=header_params)
# #     input.add_block(top=merged_header_params)
# #
# #     default_geometry_params = {'title': 'system', 'xyz_filepath': molecule.xyz.filepath}
# #     merged_geometry_params = self._merge_new_params(old_params=default_geometry_params,new_params=geometry_params)
# #     input.add_block(geometry=merged_geometry_params)
# #
# #     default_set_geometry_params = {'set_params': {'variable_name': 'geometry', 'variable_value': '\"system\"'}}
# #     merged_set_geometry_params = self._merge_new_params(old_params=default_set_geometry_params,new_params=set_geometry_params)
# #     input.add_block(top=merged_set_geometry_params)
# #
# #     default_basis_params = {'atom_strings': ('*',), 'library_strings': ('6-311G**',)}
# #     merged_basis_params = self._merge_new_params(old_params=default_basis_params,new_params=basis_params)
# #     input.add_block(basis=merged_basis_params)
# #
# #     default_dft_params = {'direct': True, 'xc_params': {'xc_functional_strings': ('b3lyp',)}}
# #     merged_dft_params = self._merge_new_params(old_params=default_dft_params,new_params=dft_params)
# #     input.add_block(dft=merged_dft_params)
# #
# #     default_dft_task_params = {'task_params': {'theory_level': 'dft', 'operation': 'energy'}}
# #     merged_dft_task_params = self._merge_new_params(old_params=default_dft_task_params,new_params=dft_task_params)
# #     input.add_block(top=merged_dft_task_params)
# #
# #     if x:
# #         default_rttddft_x_params = {'tmax':500.0,'dt':0.1,'tag':'kickx','print_dipole':True,
# #                                     'load_type': 'vectors', 'load_from': os.path.join(self.scratch_directory,'permanent','molecule.movecs'),
# #                                     'field_params':{'field_name':'kick','shape':'delta','polarization':'x','max': 1e-4},
# #                                     'excite_params':{'geometry_name':'system','field_name':'kick'}}
# #         merged_rttddft_x_params = self._merge_new_params(old_params=default_rttddft_x_params,new_params=rttddft_x_params)
# #         input.add_block(rttddft=merged_rttddft_x_params)
# #
# #         default_rttddft_x_task_params = {'task_params': {'theory_level': 'dft', 'operation': 'rt_tddft'}}
# #         merged_rttddft_x_task_params = self._merge_new_params(old_params=default_rttddft_x_task_params,new_params=rttddft_x_task_params)
# #         input.add_block(top=merged_rttddft_x_task_params)
# #
# #     if y:
# #         default_rttddft_y_params = {'tmax':500.0,'dt':0.1,'tag':'kicky','print_dipole':True,
# #                                     'load_type': 'vectors', 'load_from': os.path.join(self.scratch_directory,'permanent','molecule.movecs'),
# #                                     'field_params':{'field_name':'kick','shape':'delta','polarization':'y','max': 1e-4},
# #                                     'excite_params':{'geometry_name':'system','field_name':'kick'}}
# #         merged_rttddft_y_params = self._merge_new_params(old_params=default_rttddft_y_params,new_params=rttddft_y_params)
# #         input.add_block(rttddft=merged_rttddft_y_params)
# #
# #         default_rttddft_y_task_params = {'task_params': {'theory_level': 'dft', 'operation': 'rt_tddft'}}
# #         merged_rttddft_y_task_params = self._merge_new_params(old_params=default_rttddft_y_task_params,new_params=rttddft_y_task_params)
# #         input.add_block(top=merged_rttddft_y_task_params)
# #
# #     if z:
# #         default_rttddft_z_params = {'tmax':500.0,'dt':0.1,'tag':'kickz','print_dipole':True,
# #                                     'load_type': 'vectors', 'load_from': os.path.join(self.scratch_directory,'permanent','molecule.movecs'),
# #                                     'field_params':{'field_name':'kick','shape':'delta','polarization':'z','max': 1e-4},
# #                                     'excite_params':{'geometry_name':'system','field_name':'kick'}}
# #         merged_rttddft_z_params = self._merge_new_params(old_params=default_rttddft_z_params,new_params=rttddft_z_params)
# #         input.add_block(rttddft=merged_rttddft_z_params)
# #
# #         default_rttddft_z_task_params = {'task_params': {'theory_level': 'dft', 'operation': 'rt_tddft'}}
# #         merged_rttddft_z_task_params = self._merge_new_params(old_params=default_rttddft_z_task_params,new_params=rttddft_z_task_params)
# #         input.add_block(top=merged_rttddft_z_task_params)
# #
# #     input.write_input(input_path)
# #
# #     self.execute(input_path=input_path,output_path=output_path,num_cores=num_cores,parallel=parallel)
# #
# #     output = NWChemOutput(filename=output_path)
# #
# #     tddipole_dict = output.get_tddipole()
# #
# #     molecule.tddipole_x = (tddipole_dict['time'],tddipole_dict['mu_x'])
# #     molecule.tddipole_y = (tddipole_dict['time'],tddipole_dict['mu_y'])
# #     molecule.tddipole_z = (tddipole_dict['time'],tddipole_dict['mu_z'])



from __future__ import annotations
from typing import Any, Callable, Iterable, Optional, Tuple, Union

import cclib
import contextlib
import copy
import dlib
import functools
import materia as mtr
from materia.utils import memoize
from materia.workflow import Workflow
import numpy as np

# FIXME: replace os with pathlib where possible
import os
import pathlib
import re

# import scipy.optimize
import subprocess
import tempfile
import uuid

__all__ = [
    "ExternalTask",
    "FunctionTask",
    "InputTask",
    "MaxLIPOTR",
    "ShellCommand",
    "Task",
    "FragItFragmentStructure",
    "MultiwfnNTO",
    "MultiwfnTotalESP",
    "MultiwfnVolume",
    "PackmolSolvate",
    #    "ExecuteQChem",
    "QChemAIMD",
    "QChemKoopmanError",
    "QChemKoopmanErrorLPSCF",
    "QChemLPSCF",
    "QChemLPSCFRS",
    "QChemLRTDDFT",
    "QChemMinimizeKoopmanError",
    "QChemMinimizeKoopmanErrorLPSCF",
    "QChemOptimize",
    "QChemPolarizability",
    #    "QChemRTTDDFT",
    "QChemSinglePoint",
    "QChemSinglePointFrontier",
    "WriteQChemInput",
    "WriteQChemInputGeometryRelaxation",
    "WriteQChemInputLRTDDFT",
    "WriteQChemInputPolarizability",
    "WriteQChemInputSinglePoint",
    #    "WriteQChemTDSCF",
    "XTBOptimize",
]


class Task:
    def __init__(
        self,
        num_cores: Optional[int] = 1,
        handlers: Optional[Iterable[mtr.Handler]] = None,
        name: Optional[str] = None,
    ) -> None:
        self.num_cores = num_cores
        self.handlers = handlers or []
        self.name = (
            name
            or re.match("<class '(?P<cls>.*)'>", str(self.__class__))
            .group("cls")
            .rsplit(".")[-1]
        )

        self.requirements = []
        self.named_requirements = {}

    def requires(self, *args: Task, **kwargs: Task) -> None:
        self.requirements += [a if isinstance(a, Task) else InputTask(a) for a in args]
        self.named_requirements = dict(
            **self.named_requirements,
            **{
                k: v if isinstance(v, Task) else InputTask(v) for k, v in kwargs.items()
            },
        )

    def run(self, **kwargs: Any) -> Any:
        raise NotImplementedError

    def __str__(self) -> str:
        return self.name


class ExternalTask(Task):
    def __init__(
        self,
        engine: mtr.Engine,
        io: mtr.IO,
        handlers: Optional[Iterable[mtr.Handler]] = None,
        name: Optional[str] = None,
    ) -> None:
        self.engine = engine
        self.io = io
        super().__init__(self.engine.num_processors or 1, handlers=handlers, name=name)


class FunctionTask(Task):
    def __init__(
        self,
        f: Callable,
        handlers: Optional[Iterable[Handler]] = None,
        name: Optional[str] = None,
    ) -> None:
        super().__init__(handlers=handlers, name=name)
        self.f = f

    def run(self, **kwargs) -> None:
        return self.f(**kwargs)


class InputTask(Task):
    def __init__(
        self,
        value: Any,
        handlers: Optional[Iterable[Handler]] = None,
        name: Optional[str] = None,
    ) -> None:
        super().__init__(1, handlers=handlers, name=name)
        self.value = value

    def run(self, *args) -> Any:
        return self.value


class ShellCommand(Task):
    def __init__(
        self,
        command: str,
        handlers: Optional[Iterable[Handler]] = None,
        name: Optional[str] = None,
    ) -> None:
        super().__init__(handlers=handlers, name=name)
        self.command = command

    def run(self) -> None:
        subprocess.call(self.command.split())


def task(
    f: Callable = None,
    handlers: Optional[Iterable[Handler]] = None,
    name: Optional[str] = None,
) -> FunctionTask:
    # FIXME: this is incomptabile with mtr.Workflow.run(thread=False) (i.e. with multiprocessing) because FunctionTask cannot be serialized!
    if f is None:
        return functools.partial(task, handlers=handlers, name=name)

    return FunctionTask(f=f, handlers=handlers, name=name)


# class GoldenSectionSearch:
#     def __init__(self, objective_function):
#         self.objective_function = objective_function

#     @memoize
#     def evaluate_objective(self, x):
#         return self.objective_function(x)

#     def optimize(self, delta, tolerance):
#         bracket = self._find_gss_bracket(delta=delta)

#         return scipy.optimize.minimize_scalar(
#             fun=self.evaluate_objective, bracket=bracket, method="Golden", tol=tolerance
#         )

#     def _find_gss_bracket(self, delta):
#         # FIXME: ugly but it works...
#         phi = (1 + math.sqrt(5)) / 2

#         self.evaluate_objective(x=1e-3)
#         self.evaluate_objective(x=delta)

#         while self.evaluate_objective.cache.last_result(
#             n=1
#         ) <= self.evaluate_objective.cache.last_result(n=2):
#             _, last1 = self.evaluate_objective.cache.last_args(n=1)
#             _, last2 = self.evaluate_objective.cache.last_args(n=2)

#             self.evaluate_objective(x=last1["x"] + phi * (last1["x"] - last2["x"]))

#         if len(self.evaluate_objective.cache) > 2:
#             _, last3 = self.evaluate_objective.cache.last_args(n=3)
#             _, last1 = self.evaluate_objective.cache.last_args(n=1)
#             return (last3["x"], last1["x"])
#         else:
#             _, last2 = self.evaluate_objective.cache.last_args(n=2)
#             _, last1 = self.evaluate_objective.cache.last_args(n=1)
#             return (last2["x"], last1["x"])

#         return bracket

#     # def plot_results(self):
#     #     x, y = zip(*sorted(self.evaluate_objective.cache.items()))
#     #     plt.plot(x, y)
#     #     plt.show()


T = Union[int, float]


class MaxLIPOTR(Task):
    def __init__(
        self,
        objective_function: Callable[T, T],
        handlers: Optional[Iterable[mtr.Handler]] = None,
        name: Optional[str] = None,
    ) -> None:
        super().__init__(1, handlers, name)
        self.objective_function = objective_function

    @memoize
    def _evaluate_objective(self, *args: T) -> T:
        return self.objective_function(*args)

    def run(
        self,
        x_min: Union[T, Iterable[T]],
        x_max: Union[T, Iterable[T]],
        num_evals: int,
        epsilon: Optional[float] = 0,
    ) -> Tuple[T, Union[int, float]]:

        return dlib.find_min_global(
            self._evaluate_objective,
            x_min if isinstance(x_min, list) else [x_min],
            x_max if isinstance(x_max, list) else [x_max],
            num_evals,
            solver_epsilon=epsilon,
        )

    # def plot_results(self):
    #     x, y = zip(*sorted(self.evaluate_objective.cache.items()))
    #     plt.plot(x, y)
    #     plt.show()


# -------------------------- CCDC -------------------------- #

# from __future__ import annotations
# import os
# import mtr
# import textwrap
# from typing import Iterable, Optional

# from ...workflow.tasks.task import Task

# __all__ = ["CCDCUnitCellStructure"]


# class CCDCUnitCellStructure(Task):
#     def __init__(
#         self,
#         input_name: str,
#         molecule_name: str,
#         engine: mtr.CCDCEngine,
#         handlers: Optional[Iterable[mtr.Handler]] = None,
#         name: Optional[str] = None,
#     ) -> None:
#         self.input_name = input_name
#         self.engine = engine
#         name = "'" + molecule_name + "'"
#         self.input_string = textwrap.dedent(
#             f"""\
#                     import numpy as np
#                     import ccdc.search
#                     tns = ccdc.search.TextNumericSearch()
#                     tns.add_compound_name({name})
#                     try:
#                         hit = next(hit for hit in tns.search() if hit.entry.chemical_name == {name})
#                         print('atomic_symbols: {{0}}\\n'.format(tuple(a.atomic_symbol for a in hit.molecule.atoms)))
#                         print('coordinates: {{0}}\\n'.format(np.vstack([a.coordinates for a in hit.molecule.atoms])))
#                         print('spacegroup_number_and_setting: {{0}}\\n'.format(hit.crystal.spacegroup_number_and_setting))
#                         print('cell_lengths: {{0}}\\n'.format(np.array(hit.crystal.cell_lengths)))
#                         print('cell_angles: {{0}}\\n'.format(np.array(hit.crystal.cell_angles)))
#                     except StopIteration:
#                         pass
#                     """
#         )

#     def run(self):
#         # FIXME: catch StopIteration in CCDC code
#         # FIXME: generalize past 2019 version of CCDC code
#         # FIXME: add more rigorous/better checking for CCDC crystal matches

#         # FIXME: where does input_path come from? must include self.input_name but also must include engine.work_dir...
#         mtr.CCDCInput(ccdc_script=self.input_string).write(self.input_path)

#         self.engine.execute(input_path=self.input_path)

# -------------------------- FRAGIT -------------------------- #


class FragItFragmentStructure(ExternalTask):
    """
    Task to fragment a structure using FragIt.

    Attributes:
        engine (mtr.FragItEngine): Engine which will be used to fragment structure.
    """

    def run(
        self, structure: Union[str, mtr.Structure]
    ) -> Tuple[str, Iterable[mtr.Structure]]:
        with self.io() as io:
            structure.write(io.inp)

            self.engine.execute(self.io)
            name = pathlib.Path(io.inp).stem

            return tuple(
                mtr.Structure.read(str(p))
                for p in pathlib.Path(io.work_dir).glob(f"{name}_fragment_*.xyz")
            )


# -------------------------- MULTIWFN -------------------------- #


class MultiwfnBaseTask(ExternalTask):
    def commands(self) -> Iterable[Union[str, int, float]]:
        raise NotImplementedError

    def parse(self, output: str) -> Any:
        raise NotImplementedError

    def run(self, filepath: str) -> Any:
        inp = mtr.MultiwfnInput(mtr.expand(filepath), *self.commands(), -10)

        with self.io() as io:
            inp.write(io.inp)

            self.engine.execute(self.io)

            return self.parse(io.out)


# class ExecuteMultiwfn(Task):
#     def __init__(
#         self,
#         input_path: str,
#         engine: mtr.MultiwfnEngine,
#         handlers: Optional[Iterable[mtr.Handler]] = None,
#         name: Optional[str] = None,
#     ) -> None:
#         super().__init__(handlers=handlers, name=name)
#         self.input_path = input_path
#         self.engine = engine

#     def run(self) -> None:
#         self.engine.execute(input_path=self.input_path)


# excitations = mtr.QChemOutput(filepath=filepath).get("electronic_excitations")
# excitation_filepath = mtr.expand(f"{io.work_dir}/excitations.txt")
# with open(excitation_filepath, "w") as f:
#     f.write(excitations.to_gaussian())

# inp = mtr.MultiwfnInput(mtr.expand(filepath), *self.commands(excitation_filepath), -10)
# inp.write(io.inp)


class MultiwfnNTO(MultiwfnBaseTask):
    def commands(
        self, excitation_filepath: str, work_dir: str
    ) -> Tuple[Union[str, int, float]]:
        return (
            18,
            *(
                a
                for i in range(40)
                for a in [6]
                + ([excitation_filepath] if i == 0 else [])
                + [i + 1, 2, mtr.expand(f"{work_dir}/S{i+1}.fch")]
            ),
            0,
        )

    def parse(self, output: str) -> None:
        # FIXME: is there some useful output to return?
        return None

    def run(self, filepath: str, excitation_filepath: str) -> Any:
        with self.io() as io:
            inp = mtr.MultiwfnInput(
                mtr.expand(filepath),
                *self.commands(excitation_filepath, io.work_dir),
                -10,
            )
            inp.write(io.inp)

            self.engine.execute(self.io)

            return self.parse(io.out)


class MultiwfnTotalESP(MultiwfnBaseTask):
    def __init__(
        self,
        grid_quality: str,
        engine: mtr.Engine,
        io: mtr.IO,
        handlers: Optional[Iterable[mtr.Handler]] = None,
        name: Optional[str] = None,
    ) -> None:
        self.grid_quality = grid_quality
        super().__init__(engine=engine, io=io, handlers=handlers, name=name)

    def commands(self):
        grid_quality = dict(low=1, medium=2, high=3)
        # FIXME: finish implementing this
        raise NotImplementedError
        return (
            5,
            12,
            grid_quality[self.grid_quality],
            "0,0,0",
            0,
        )

    def parse(self, output: str) -> mtr.Qty:
        # FIXME: finish implementing this
        raise NotImplementedError
        return mtr.MultiwfnOutput(output).get("volume")


class MultiwfnVolume(MultiwfnBaseTask):
    def commands(self):
        integration_mesh_exp = 9  #: int = 9,
        density_isosurface = 1e-3  #: float = 0.001,
        box_size_factor = 1.7  #: float = 1.7
        return (
            100,
            3,
            f"{integration_mesh_exp},{density_isosurface},{box_size_factor}",
            "0,0,0",
            0,
        )

    def parse(self, output: str) -> mtr.Qty:
        return mtr.MultiwfnOutput(output).get("volume")


# class WriteMultiwfnInput(Task):
#     def __init__(
#         self,
#         input_name: str,
#         in_filepath: str,  # FIXME: awful name for this variable, fix here and analogous issues throughout this file
#         commands: Iterable[str],
#         work_directory: str = ".",
#         handlers: Iterable[Handler] = None,
#         name: str = None,
#     ):
#         super().__init__(handlers=handlers, name=name)
#         self.input_path = mtr.expand(os.path.join(work_directory, input_name))
#         self.in_filepath = mtr.expand(in_filepath)
#         self.commands = commands

#         try:
#             os.makedirs(mtr.expand(work_directory))
#         except FileExistsError:
#             pass

#     def run(self):
#         mtr.MultiwfnInput(self.in_filepath, *self.commands).write(self.input_path)


# -------------------------- OPENBABEL -------------------------- #

# from __future__ import annotations
# import contextlib
# import os
# import mtr
# import re
# import tempfile
# from typing import Iterable, Optional, Union

# from ...workflow.tasks.task import Task

# __all__ = [
#     "OpenbabelConvertToFile",
#     "OpenbabelConvertToMol",
#     "OpenbabelConvertToPDB",
#     "OpenbabelConvertToSMILES",
#     "OpenbabelConvertToSDF",
# ]


# class OpenbabelConvertToFile(Task):
#     def __init__(
#         self,
#         engine: mtr.OpenbabelEngine,
#         filetype: str,
#         output_name: Optional[str] = None,
#         log_name: Optional[str] = "obabel.log",
#         work_dir: Optional[str] = ".",
#         keep_logs: bool = True,
#         handlers: Optional[Iterable[mtr.Handler]] = None,
#         name: Optional[str] = None,
#     ) -> None:
#         super().__init__(handlers=handlers, name=name)
#         self.engine = engine
#         self.filetype = filetype
#         self.output_name = output_name
#         self.log_name = log_name
#         self.work_dir = work_dir
#         self.keep_logs = keep_logs

#     def run(self, structure: Union[str, mtr.Structure]) -> str:
#         with mtr.work_dir(self.work_dir) as wd:
#             arguments = [f"-o{self.filetype}"]

#             if self.output_name is not None:
#                 output_path = (
#                     f"{mtr.expand(self.output_name,dir=wd)}.{self.filetype}"
#                 )
#                 arguments.append(f"-O{output_path}")

#             with contextlib.nullcontext(structure) if isinstance(
#                 structure, str
#             ) else structure.tempfile(suffix=".xyz") as fp:
#                 input_filepath = mtr.expand(
#                     path=fp.name if hasattr(fp, "name") else fp, dir=wd
#                 )
#                 self.engine.execute(
#                     input_filepath=input_filepath,
#                     log_filepath=mtr.expand(path=self.log_name, dir=wd),
#                     arguments=arguments,
#                 )

#             if self.output_name is not None:
#                 with open(output_path, "r") as f:
#                     return "".join(f.readlines())


# class OpenbabelConvertToMol(OpenbabelConvertToFile):
#     def __init__(
#         self,
#         engine: mtr.OpenbabelEngine,
#         output_name: str,
#         log_name: Optional[str] = "obabel.log",
#         work_dir: Optional[str] = ".",
#         keep_logs: bool = True,
#         handlers: Optional[Iterable[mtr.Handler]] = None,
#         name: Optional[str] = None,
#     ) -> None:
#         super().__init__(
#             engine=engine,
#             filetype="mol",
#             output_name=output_name,
#             log_name=log_name,
#             work_dir=work_dir,
#             keep_logs=keep_logs,
#             handlers=handlers,
#             name=name,
#         )


# class OpenbabelConvertToPDB(OpenbabelConvertToFile):
#     def __init__(
#         self,
#         engine: mtr.OpenbabelEngine,
#         output_name: str,
#         log_name: Optional[str] = "obabel.log",
#         work_dir: Optional[str] = ".",
#         keep_logs: bool = True,
#         handlers: Optional[Iterable[mtr.Handler]] = None,
#         name: Optional[str] = None,
#     ) -> None:
#         super().__init__(
#             engine=engine,
#             filetype="pdb",
#             output_name=output_name,
#             log_name=log_name,
#             work_dir=work_dir,
#             keep_logs=keep_logs,
#             handlers=handlers,
#             name=name,
#         )


# class OpenbabelConvertToSDF(OpenbabelConvertToFile):
#     def __init__(
#         self,
#         engine: mtr.OpenbabelEngine,
#         output_name: str,
#         log_name: Optional[str] = "obabel.log",
#         work_dir: Optional[str] = ".",
#         keep_logs: bool = True,
#         handlers: Optional[Iterable[mtr.Handler]] = None,
#         name: Optional[str] = None,
#     ) -> None:
#         super().__init__(
#             engine=engine,
#             filetype="sdf",
#             output_name=output_name,
#             log_name=log_name,
#             work_dir=work_dir,
#             keep_logs=keep_logs,
#             handlers=handlers,
#             name=name,
#         )


# class OpenbabelConvertToSMILES(OpenbabelConvertToFile):
#     def __init__(
#         self,
#         engine: mtr.OpenbabelEngine,
#         output_name: str,
#         log_name: Optional[str] = "obabel.log",
#         work_dir: Optional[str] = ".",
#         keep_logs: bool = True,
#         handlers: Optional[Iterable[mtr.Handler]] = None,
#         name: Optional[str] = None,
#     ) -> None:
#         super().__init__(
#             engine=engine,
#             filetype="smi",
#             output_name=output_name,
#             log_name=log_name,
#             work_dir=work_dir,
#             keep_logs=keep_logs,
#             handlers=handlers,
#             name=name,
#         )

#     def run(self, structure: Union[str, mtr.Structure]) -> str:
#         super().run(structure=structure)

#         with open(
#             f"{mtr.expand(self.output_name,dir=self.work_dir)}.smi", "r"
#         ) as f:
#             smiles, *_ = re.search(r"([^\s]*).*", "".join(f.readlines())).groups()

#         return smiles


# -------------------------- PACKMOL -------------------------- #


class PackmolSolvate(ExternalTask):
    def _packing_params(
        self, shells: int, number_density: Optional[mtr.Qty] = None
    ) -> Tuple[int, mtr.Qty]:
        # these are the ideal gas packing values:
        n = int((2 / 3) * shells ** 3)
        sphere_radius = shells * (2 * np.pi * number_density) ** (-1 / 3)

        return n, sphere_radius

    def run(
        self,
        solute: Union[mtr.Structure, str],
        solvent: Union[mtr.Structure, str],
        shells: int,
        tolerance: float,
        solvent_density: mtr.Qty,
    ) -> mtr.Structure:
        if solvent_density.dimension == mtr.Dimension(M=1, L=-3):
            number_density = solvent_density / solvent.mass
        else:
            number_density = solvent_density

        n, sphere_radius = self._packing_params(
            shells=shells, number_density=number_density
        )

        with self.io() as io:
            inp = mtr.PackmolInput(
                tolerance=tolerance,
                filetype="xyz",
                output_name=mtr.expand(path="packed", dir=io.work_dir),
            )

            if isinstance(solute, str):
                solute_cm = contextlib.nullcontext(solute)
            else:
                solute_cm = solute.tempfile(suffix=".xyz", dir=io.work_dir)

            if isinstance(solvent, str):
                solvent_cm = contextlib.nullcontext(solvent)
            else:
                solvent_cm = solvent.tempfile(suffix=".xyz", dir=io.work_dir)

            with solute_cm as f, solvent_cm as g:
                inp.add_structure(
                    structure_filepath=mtr.expand(
                        path=f.name if hasattr(f, "name") else f, dir=io.work_dir
                    ),
                    number=1,
                    instructions=["fixed 0. 0. 0. 0. 0. 0."],
                )

                inp.add_structure(
                    structure_filepath=mtr.expand(
                        path=g.name if hasattr(g, "name") else g, dir=io.work_dir
                    ),
                    number=n - 1,
                    instructions=[
                        f"inside sphere 0. 0. 0. {sphere_radius.convert(mtr.angstrom).value}"
                    ],
                )

                inp.write(io.inp)

                self.engine.execute(self.io)

                return mtr.Structure.read(
                    mtr.expand(path="packed.xyz", dir=io.work_dir)
                )


# -------------------------- QCHEM -------------------------- #


class QChemBaseTask(ExternalTask):
    def defaults(self, settings: mtr.Settings) -> mtr.Settings:
        raise NotImplementedError

    def parse(self, output: str) -> Any:
        raise NotImplementedError

    def run(
        self,
        structure: Union[mtr.QChemStructure, mtr.QChemFragments, mtr.Structure],
        settings: Optional[mtr.Settings] = None,
        arguments: Optional[Iterable[str]] = None,
    ) -> Any:
        s = mtr.Settings() if settings is None else copy.deepcopy(settings)

        # FIXME: this is essentially a hotpatch to handle fragments - come up with something more elegant/sensible ASAP
        inp = mtr.QChemInput(
            molecule=structure
            if isinstance(structure, mtr.Structure)
            or isinstance(structure, mtr.QChemStructure)
            or isinstance(structure, mtr.QChemFragments)
            else mtr.QChemFragments(structures=structure),
            settings=self.defaults(s),
        )

        with self.io() as io:
            inp.write(io.inp)

            self.engine.execute(self.io, arguments=arguments)

            return self.parse(io.out)


# class ExecuteQChem(QChemBaseTask):
#     def run(self) -> Any:
#         with self.io() as io:
#             self.engine.execute(io.inp, io.out)


class QChemAIMD(QChemBaseTask):
    def parse(self, output: str) -> Any:
        with open(output, "r") as f:
            return "".join(f.readlines())

    def defaults(self, settings: mtr.Settings) -> mtr.Settings:
        if ("rem", "exchange") not in settings and ("rem", "method",) not in settings:
            settings["rem", "exchange"] = "HF"
        if ("rem", "basis") not in settings:
            settings["rem", "basis"] = "3-21G"
        if ("rem", "jobtype") not in settings:
            settings["rem", "jobtype"] = "aimd"
        if ("rem", "time_step") not in settings:
            settings["rem", "time_step"] = 1
        if ("rem", "aimd_steps") not in settings:
            settings["rem", "aimd_steps"] = 10
        if ("velocity",) not in settings and (
            "rem",
            "aimd_init_veloc",
        ) not in settings:
            settings["rem", "aimd_init_veloc"] = "thermal"
        if (
            ("rem", "aimd_init_veloc") in settings
            and settings["rem", "aimd_init_veloc"].lower().strip() == "thermal"
            and ("rem", "aimd_temp") not in settings
        ):
            settings["rem", "aimd_temp"] = 300

        return settings


class QChemLPSCF(QChemBaseTask):
    def parse(self, output: str) -> Any:
        try:
            energy = cclib.io.ccread(output).scfenergies * mtr.eV
        except AttributeError:
            energy = None

        return energy

    def defaults(self, settings: mtr.Settings) -> mtr.Settings:
        if ("rem", "exchange") not in settings and ("rem", "method",) not in settings:
            settings["rem", "exchange"] = "HF"
        if ("rem", "basis") not in settings:
            settings["rem", "basis"] = "3-21G"
        if ("rem", "jobtype") not in settings:
            settings["rem", "jobtype"] = "sp"
        if ("rem", "frgm_method") not in settings:
            settings["rem", "frgm_method"] = "stoll"
        if ("rem_frgm", "scf_convergence") not in settings:
            settings["rem_frgm", "scf_convergence"] = 2
        if ("rem_frgm", "thresh") not in settings:
            settings["rem_frgm", "thresh"] = 5

        return settings

    def run(
        self,
        *fragments: Union[mtr.QChemStructure, mtr.Structure],
        settings: Optional[mtr.Settings] = None,
    ) -> Any:
        s = mtr.Settings() if settings is None else copy.deepcopy(settings)

        # FIXME: this is essentially a hotpatch to handle fragments - come up with something more elegant/sensible ASAP
        inp = mtr.QChemInput(
            molecule=mtr.QChemFragments(structures=fragments),
            settings=self.defaults(s),
        )

        with self.io() as io:
            inp.write(io.inp)

            self.engine.execute(self.io)

            return self.parse(io.out)


class QChemLPSCFRS(QChemBaseTask):
    def defaults(self, settings: mtr.Settings) -> mtr.Settings:
        if ("rem", "exchange") not in settings and ("rem", "method",) not in settings:
            settings["rem", "exchange"] = "HF"
        if ("rem", "basis") not in settings:
            settings["rem", "basis"] = "3-21G"
        if ("rem", "jobtype") not in settings:
            settings["rem", "jobtype"] = "sp"
        if ("rem", "frgm_method") not in settings:
            settings["rem", "frgm_method"] = "stoll"
        if ("rem_frgm", "scf_convergence") not in settings:
            settings["rem_frgm", "scf_convergence"] = 2
        if ("rem_frgm", "thresh") not in settings:
            settings["rem_frgm", "thresh"] = 5
        if ("rem_frgm", "frgm_lpcorr") not in settings:
            settings["rem_frgm", "frgm_lpcorr"] = "rs_exact_scf"

        return settings


class QChemKoopmanError(Task):
    def __init__(
        self,
        engine: mtr.Engine,
        gs_io: mtr.IO,
        cation_io: mtr.IO,
        anion_io: mtr.IO,
        handlers: Optional[Iterable[mtr.Handler]] = None,
        name: Optional[str] = None,
    ) -> None:
        super().__init__(
            (engine.num_threads or 1) * (engine.num_processors or 1),
            handlers=handlers,
            name=name,
        )
        self.engine = engine
        self.gs_io = gs_io
        self.cation_io = cation_io
        self.anion_io = anion_io

    def defaults(self, settings: mtr.Settings) -> mtr.Settings:
        if ("rem", "exchange") not in settings and ("rem", "method") not in settings:
            settings["rem", "exchange"] = "hf"
        if ("rem", "basis") not in settings:
            settings["rem", "basis"] = "3-21G"
        if ("rem", "jobtype") not in settings:
            settings["rem", "jobtype"] = "sp"

        return settings

    def run(
        self,
        structure: Union[mtr.QChemStructure, mtr.QChemFragments, mtr.Structure],
        settings: Optional[mtr.Settings] = None,
        num_consumers: Optional[int] = 1,
    ) -> float:
        s = mtr.Settings() if settings is None else copy.deepcopy(settings)

        # if ("structure", "charge") not in settings:
        #     settings["structure", "charge"] = 0
        #     # FIXME: rather than guessing 0, use rdkit.Chem.rdmolops.GetFormalCharge?
        # if ("structure", "multiplicity") not in settings:
        #     settings["structure", "multiplicity"] = 1

        input_settings = self.defaults(s)

        gs = QChemSinglePointFrontier(self.engine, self.gs_io, name="gs")
        gs_structure = mtr.QChemStructure(structure, charge=0, multiplicity=1)
        gs.requires(structure=gs_structure, settings=input_settings)

        cation = QChemSinglePoint(self.engine, self.cation_io, name="cation")
        cation_structure = mtr.QChemStructure(structure, charge=1, multiplicity=2)
        cation.requires(structure=cation_structure, settings=input_settings)

        anion = QChemSinglePoint(self.engine, self.anion_io, name="anion")
        anion_structure = mtr.QChemStructure(structure, charge=-1, multiplicity=2)
        anion.requires(structure=anion_structure, settings=input_settings)

        wf = Workflow(gs, cation, anion)

        out = wf.run(available_cores=self.num_cores, num_consumers=num_consumers)

        energy, homo, lumo = out["gs"]
        cation = out["cation"]
        anion = out["anion"]

        ea = energy - anion
        ip = cation - energy

        J_squared = (ea + lumo) ** 2 + (ip + homo) ** 2

        return np.sqrt(J_squared.convert(mtr.eV ** 2).value.item())


class QChemKoopmanErrorLPSCF(Task):
    def __init__(
        self,
        engine: mtr.Engine,
        gs_io: mtr.IO,
        cation_io: mtr.IO,
        anion_io: mtr.IO,
        handlers: Optional[Iterable[mtr.Handler]] = None,
        name: Optional[str] = None,
    ) -> None:
        super().__init__(
            (engine.num_threads or 1) * (engine.num_processors or 1),
            handlers=handlers,
            name=name,
        )
        self.engine = engine
        self.gs_io = gs_io
        self.cation_io = cation_io
        self.anion_io = anion_io

    def defaults(self, settings: mtr.Settings) -> mtr.Settings:
        if ("rem", "exchange") not in settings and ("rem", "method") not in settings:
            settings["rem", "exchange"] = "hf"
        if ("rem", "basis") not in settings:
            settings["rem", "basis"] = "3-21G"
        if ("rem", "jobtype") not in settings:
            settings["rem", "jobtype"] = "sp"

        return settings

    def run(
        self,
        *fragments: mtr.Structure,
        active_fragment: int,
        settings: Optional[mtr.Settings] = None,
        num_consumers: Optional[int] = 1,
    ) -> float:
        s = mtr.Settings() if settings is None else copy.deepcopy(settings)

        # if ("structure", "charge") not in settings:
        #     settings["structure", "charge"] = 0
        #     # FIXME: rather than guessing 0, use rdkit.Chem.rdmolops.GetFormalCharge?
        # if ("structure", "multiplicity") not in settings:
        #     settings["structure", "multiplicity"] = 1

        gs_charges = [0] * len(fragments)
        gs_multiplicities = [1] * len(fragments)

        cation_charges = gs_charges.copy()
        cation_multiplicities = gs_multiplicities.copy()
        cation_charges[active_fragment] += 1
        cation_multiplicities[active_fragment] = 2

        anion_charges = gs_charges.copy()
        anion_multiplicities = gs_multiplicities.copy()
        anion_charges[active_fragment] -= 1
        anion_multiplicities[active_fragment] = 2

        input_settings = self.defaults(s)

        gs = QChemSinglePointFrontier(self.engine, self.gs_io, name="gs")
        gs_structure = mtr.QChemFragments(
            fragments,
            gs_charges,
            gs_multiplicities,
            total_charge=sum(gs_charges),
            total_multiplicity=1,
        )
        gs.requires(structure=gs_structure, settings=input_settings)

        cation = QChemSinglePoint(self.engine, self.cation_io, name="cation")
        cation_structure = mtr.QChemFragments(
            fragments,
            cation_charges,
            cation_multiplicities,
            total_charge=sum(cation_charges),
            total_multiplicity=2,
        )
        cation.requires(structure=cation_structure, settings=input_settings)

        anion = QChemSinglePoint(self.engine, self.anion_io, name="anion")
        anion_structure = mtr.QChemFragments(
            fragments,
            anion_charges,
            anion_multiplicities,
            total_charge=sum(anion_charges),
            total_multiplicity=2,
        )
        anion.requires(structure=anion_structure, settings=input_settings)

        wf = Workflow(gs, cation, anion)

        out = wf.run(available_cores=self.num_cores, num_consumers=num_consumers)

        energy, homo, lumo = out["gs"]
        cation = out["cation"]
        anion = out["anion"]

        ea = energy - anion
        ip = cation - energy

        J_squared = (ea + lumo) ** 2 + (ip + homo) ** 2

        return np.sqrt(J_squared.convert(mtr.eV ** 2).value.item())


class QChemLRTDDFT(QChemBaseTask):
    def parse(self, output: str) -> Any:
        return mtr.QChemOutput(filepath=output).get("electronic_excitations")

    def defaults(self, settings: mtr.Settings) -> mtr.Settings:
        if ("rem", "exchange") not in settings and ("rem", "method",) not in settings:
            settings["rem", "exchange"] = "HF"
        if ("rem", "basis") not in settings:
            settings["rem", "basis"] = "3-21G"
        if ("rem", "cis_n_roots") not in settings:
            settings["rem", "cis_n_roots"] = 1
        if ("rem", "cis_singlets") not in settings:
            settings["rem", "cis_singlets"] = True
        if ("rem", "cis_triplets") not in settings:
            settings["rem", "cis_triplets"] = False
        if ("rem", "rpa") not in settings:
            settings["rem", "rpa"] = False

        return settings


class QChemMinimizeKoopmanError(Task):
    def __init__(
        self,
        engine: mtr.Engine,
        io: mtr.IO,
        handlers: Optional[Iterable[mtr.Handler]] = None,
        name: Optional[str] = None,
    ) -> None:
        super().__init__(
            (engine.num_threads or 1) * (engine.num_processors or 1),
            handlers=handlers,
            name=name,
        )
        self.engine = engine
        self.io = io

    def defaults(self, settings: mtr.Settings) -> mtr.Settings:
        if ("rem", "basis") not in settings:
            settings["rem", "basis"] = "3-21G"
        if ("rem", "jobtype") not in settings:
            settings["rem", "jobtype"] = "sp"
        if ("rem", "exchange") not in settings:
            settings["rem", "exchange"] = "gen"
        if ("rem", "lrc_dft") not in settings:
            settings["rem", "lrc_dft"] = True
        if ("rem", "src_dft") not in settings:
            settings["rem", "src_dft"] = 2

        return settings

    def run(
        self,
        structure: Union[mtr.QChemStructure, mtr.QChemFragments, mtr.Structure],
        settings: Optional[mtr.Settings] = None,
        epsilon: Optional[Union[int, float]] = 1.0,
        alpha: Optional[float] = None,
        num_evals: Optional[int] = 5,
    ) -> float:
        def _objective(omega, _alpha):
            beta = 1 / epsilon - _alpha

            s = self.defaults(settings)
            s["rem", "hf_sr"] = int(round(1000 * _alpha))
            s["rem", "hf_lr"] = int(round(1000 * (_alpha + beta)))
            s["xc_functional"] = (
                ("X", "HF", _alpha),
                ("X", "wPBE", beta),
                ("X", "PBE", 1 - _alpha - beta),
                ("C", "PBE", 1.0),
            )
            omega = int(round(1000 * omega))
            s["rem", "omega"] = s["rem", "omega2"] = omega

            wd = mtr.expand(f"{io.work_dir}/{omega}")

            gs_io = mtr.IO("gs.in", "gs.out", wd)
            cation_io = mtr.IO("cation.in", "cation.out", wd)
            anion_io = mtr.IO("anion.in", "anion.out", wd)

            ke = mtr.QChemKoopmanError(self.engine, gs_io, cation_io, anion_io)
            # FIXME: not sure the best way to handle num_consumers here...
            return ke.run(structure, s, num_consumers=3)

        with self.io() as io:
            if alpha is None:
                return mtr.MaxLIPOTR(_objective).run(
                    x_min=[1e-3, 0], x_max=[1, 1 / epsilon], num_evals=num_evals
                )
            else:
                return mtr.MaxLIPOTR(functools.partial(_objective, _alpha=alpha)).run(
                    x_min=1e-3, x_max=1, num_evals=num_evals
                )


class QChemMinimizeKoopmanErrorLPSCF(Task):
    def __init__(
        self,
        engine: mtr.Engine,
        io: mtr.IO,
        handlers: Optional[Iterable[mtr.Handler]] = None,
        name: Optional[str] = None,
    ) -> None:
        super().__init__(
            (engine.num_threads or 1) * (engine.num_processors or 1),
            handlers=handlers,
            name=name,
        )
        self.engine = engine
        self.io = io

    def defaults(self, settings: mtr.Settings) -> mtr.Settings:
        if ("rem", "basis") not in settings:
            settings["rem", "basis"] = "3-21G"
        if ("rem", "jobtype") not in settings:
            settings["rem", "jobtype"] = "sp"
        if ("rem", "exchange") not in settings:
            settings["rem", "exchange"] = "gen"
        if ("rem", "lrc_dft") not in settings:
            settings["rem", "lrc_dft"] = True
        if ("rem", "src_dft") not in settings:
            settings["rem", "src_dft"] = 2

        return settings

    def run(
        self,
        *fragments: mtr.Structure,
        active_fragment: int,
        settings: Optional[mtr.Settings] = None,
        epsilon: Optional[Union[int, float]] = 1.0,
        alpha: Optional[float] = 0.2,
        num_evals: Optional[int] = 5,
    ) -> float:
        beta = 1 / epsilon - alpha

        s = self.defaults(settings)
        s["rem", "hf_sr"] = int(round(1000 * alpha))
        s["rem", "hf_lr"] = int(round(1000 * (alpha + beta)))
        s["xc_functional"] = (
            ("X", "HF", alpha),
            ("X", "wPBE", beta),
            ("X", "PBE", 1 - alpha - beta),
            ("C", "PBE", 1.0),
        )

        with self.io() as io:

            def f(omega):
                omega = int(round(1000 * omega))
                s["rem", "omega"] = s["rem", "omega2"] = omega

                wd = mtr.expand(f"{io.work_dir}/{omega}")

                gs_io = mtr.IO("gs.in", "gs.out", wd)
                cation_io = mtr.IO("cation.in", "cation.out", wd)
                anion_io = mtr.IO("anion.in", "anion.out", wd)

                ke = mtr.QChemKoopmanErrorLPSCF(self.engine, gs_io, cation_io, anion_io)
                # FIXME: not sure the best way to handle num_consumers here...
                return ke.run(
                    *fragments,
                    active_fragment=active_fragment,
                    settings=settings,
                    num_consumers=3,
                )

            return mtr.MaxLIPOTR(f).run(x_min=1e-3, x_max=1, num_evals=num_evals)


class QChemOptimize(QChemBaseTask):
    def parse(self, output: str) -> Any:
        try:
            parsed = cclib.io.ccread(output)
            # FIXME: is this the correct unit?
            coords = parsed.atomcoords[-1, :, :] * mtr.angstrom
            zs = parsed.atomnos
        except AttributeError:
            return None
        # FIXME: is this the correct unit?
        atoms = (mtr.Atom(element=Z, position=p) for Z, p in zip(zs, coords))
        return mtr.Structure(*atoms)

    def defaults(self, settings: mtr.Settings) -> mtr.Settings:
        if ("rem", "exchange") not in settings and ("rem", "method",) not in settings:
            settings["rem", "exchange"] = "HF"
        if ("rem", "basis") not in settings:
            settings["rem", "basis"] = "3-21G"
        if ("rem", "jobtype") not in settings:
            settings["rem", "jobtype"] = "opt"

        return settings


class QChemPolarizability(QChemBaseTask):
    def parse(self, output: str) -> Any:
        try:
            polarizability = (
                cclib.io.ccread(output).polarizabilities[-1] * mtr.au_volume
            )
        except AttributeError:
            polarizability = None

        return polarizability

    def defaults(self, settings: mtr.Settings) -> mtr.Settings:
        if ("rem", "exchange") not in settings and ("rem", "method",) not in settings:
            settings["rem", "exchange"] = "HF"
        if ("rem", "basis") not in settings:
            settings["rem", "basis"] = "3-21G"
        if ("rem", "jobtype") not in settings:
            settings["rem", "jobtype"] = "polarizability"

        return settings

    def run(
        self,
        structure: Union[mtr.QChemStructure, mtr.QChemFragments, mtr.Structure],
        settings: Optional[mtr.Settings] = None,
    ) -> Any:
        # NOTE: bug workaround for parallel polarizability calculation in Q-Chem 5.2.1
        os.environ["QCINFILEBASE"] = "0"
        return super().run(structure, settings)


# class QChemRTTDDFT(Task):
#     def __init__(
#         self,
#         structure,
#         input_name,
#         output_name,
#         scratch_directory,
#         settings=None,
#         tdscf_settings=None,
#         executable="qchem",
#         work_directory=".",
#         num_cores=1,
#         parallel=False,
#         handlers=None,
#         name=None,
#     ):
#         super().__init__(handlers=handlers, name=name)
#         self.work_directory = mtr.expand(work_directory)
#         self.input_path = mtr.expand(os.path.join(work_directory, input_name))
#         self.output_path = mtr.expand(os.path.join(work_directory, output_name))
#         self.scratch_directory = mtr.expand(scratch_directory)
#         self.executable = executable
#         self.num_cores = num_cores
#         self.parallel = parallel
#         try:
#             os.makedirs(mtr.expand(work_directory))
#         except FileExistsError:
#             pass
#         settings = settings or mtr.Settings()
#         settings["molecule", "structure"] = structure
#         if ("rem", "exchange") not in settings and ("rem", "method",) not in settings:
#             settings["rem", "exchange"] = "HF"
#         if ("rem", "basis") not in settings:
#             settings["rem", "basis"] = "3-21G"
#         if ("rem", "rttddft") not in settings:
#             settings["rem", "rttddft"] = 1
#         self.tdscf_settings = tdscf_settings or mtr.Settings()

#     def run(self):
#         tdscf_input_path = mtr.expand(os.path.join(self.work_directory, "TDSCF.prm"))
#         keys = tuple(str(next(iter(k))) for k in self.tdscf_settings)
#         max_length = max(len(k) for k in keys)
#         with open(mtr.expand(tdscf_input_path), "w") as f:
#             f.write(
#                 "\n".join(
#                     k + " " * (max_length - len(k) + 1) + str(self.tdscf_settings[k])
#                     for k in keys
#                 )
#             )
#         mtr.QChemInput(settings=settings).write(filepath=self.input_path)
#         try:
#             os.makedirs(mtr.expand(os.path.join(self.work_directory, "logs")))
#         except FileExistsError:
#             pass
#         os.environ["QCSCRATCH"] = self.scratch_directory
#         with open(self.output_path, "w") as f:
#             if self.parallel:
#                 subprocess.call(
#                     [self.executable, "-nt", str(self.num_cores), self.input_path],
#                     stdout=f,
#                     stderr=subprocess.STDOUT,
#                 )
#             else:
#                 subprocess.call([self.executable, self.input_path], stdout=f)

#         # FIXME: finish with output


class QChemSinglePoint(QChemBaseTask):
    def parse(self, output: str) -> Any:
        try:
            energy = cclib.io.ccread(output).scfenergies * mtr.eV
        except AttributeError:
            energy = None

        return energy

    def defaults(self, settings: mtr.Settings) -> mtr.Settings:
        if ("rem", "exchange") not in settings and ("rem", "method",) not in settings:
            settings["rem", "exchange"] = "HF"
        if ("rem", "basis") not in settings:
            settings["rem", "basis"] = "3-21G"
        if ("rem", "jobtype") not in settings:
            settings["rem", "jobtype"] = "sp"

        return settings


class QChemSinglePointFrontier(QChemBaseTask):
    def parse(self, output: str) -> Any:
        try:
            out = cclib.io.ccread(output)
            energy = out.scfenergies * mtr.eV
            moenergies = out.moenergies
            homo_indices = out.homos

            homos = []
            lumos = []

            for moe, h in zip(moenergies, homo_indices):
                homo, lumo = moe[h : h + 2]
                homos.append(homo)
                lumos.append(lumo)

            homo = max(homos) * mtr.eV
            lumo = min(lumos) * mtr.eV
        except AttributeError:
            energy = None
            homo = None
            lumo = None

        return energy, homo, lumo

    def defaults(self, settings: mtr.Settings) -> mtr.Settings:
        if ("rem", "exchange") not in settings and ("rem", "method") not in settings:
            settings["rem", "exchange"] = "HF"
        if ("rem", "basis") not in settings:
            settings["rem", "basis"] = "3-21G"
        if ("rem", "jobtype") not in settings:
            settings["rem", "jobtype"] = "sp"

        return settings


class WriteQChemInput(Task):
    def __init__(
        self,
        io: mtr.IO,
        handlers: Optional[Iterable[mtr.Handler]] = None,
        name: str = None,
    ) -> None:
        super().__init__(handlers=handlers, name=name)
        self.io = io

    def defaults(self, settings: mtr.Settings) -> mtr.Settings:
        return settings

    def run(
        self,
        structure: Union[mtr.QChemStructure, mtr.QChemFragments, mtr.Structure],
        settings: Optional[mtr.Settings] = None,
    ) -> None:
        s = mtr.Settings() if settings is None else copy.deepcopy(settings)
        # FIXME: this is essentially a hotpatch to handle fragments - come up with something more elegant/sensible ASAP
        inp = mtr.QChemInput(
            molecule=structure
            if isinstance(structure, mtr.Structure)
            or isinstance(structure, mtr.QChemStructure)
            else mtr.QChemFragments(structures=structure),
            settings=self.defaults(s),
        )

        with self.io() as io:
            inp.write(io.inp)


class WriteQChemInputGeometryRelaxation(WriteQChemInput):
    def defaults(self, settings: mtr.Settings) -> mtr.Settings:
        if ("rem", "exchange") not in settings and ("rem", "method",) not in settings:
            settings["rem", "exchange"] = "HF"
        if ("rem", "basis") not in settings:
            settings["rem", "basis"] = "3-21G"
        if ("rem", "jobtype") not in settings:
            settings["rem", "jobtype"] = "opt"

        return settings


class WriteQChemInputLRTDDFT(WriteQChemInput):
    def defaults(self, settings: mtr.Settings) -> mtr.Settings:
        if ("rem", "exchange") not in settings and ("rem", "method",) not in settings:
            settings["rem", "exchange"] = "HF"
        if ("rem", "basis") not in settings:
            settings["rem", "basis"] = "3-21G"
        if ("rem", "cis_n_roots") not in settings:
            settings["rem", "cis_n_roots"] = 1
        if ("rem", "cis_singlets") not in settings:
            settings["rem", "cis_singlets"] = True
        if ("rem", "cis_triplets") not in settings:
            settings["rem", "cis_triplets"] = False
        if ("rem", "rpa") not in settings:
            settings["rem", "rpa"] = False

        return settings


class WriteQChemInputPolarizability(WriteQChemInput):
    def defaults(self, settings: mtr.Settings) -> mtr.Settings:
        if ("rem", "exchange") not in settings and ("rem", "method",) not in settings:
            settings["rem", "exchange"] = "HF"
        if ("rem", "basis") not in settings:
            settings["rem", "basis"] = "3-21G"
        if ("rem", "jobtype") not in settings:
            settings["rem", "jobtype"] = "polarizability"

        return settings


class WriteQChemInputSinglePoint(WriteQChemInput):
    def defaults(self, settings: mtr.Settings) -> mtr.Settings:
        if ("rem", "exchange") not in settings and ("rem", "method",) not in settings:
            settings["rem", "exchange"] = "HF"
        if ("rem", "basis") not in settings:
            settings["rem", "basis"] = "3-21G"

        return settings


# class WriteQChemTDSCF(Task):
#     def __init__(
#         self,
#         settings: Optional[mtr.Settings] = None,
#         work_directory: str = ".",
#         handlers: Optional[Iterable[mtr.Handler]] = None,
#         name: str = None,
#     ):
#         super().__init__(handlers=handlers, name=name)
#         self.work_directory = mtr.expand(work_directory)
#         settings = settings

#         try:
#             os.makedirs(mtr.expand(work_directory))
#         except FileExistsError:
#             pass

#     def run(self) -> None:
#         input_path = mtr.expand(os.path.join(self.work_directory, "TDSCF.prm"))

#         keys = tuple(str(next(iter(k))) for k in settings)
#         max_length = max(len(k) for k in keys)

#         with open(mtr.expand(input_path), "w") as f:
#             f.write(
#                 "\n".join(
#                     k + " " * (max_length - len(k) + 1) + str(settings[k]) for k in keys
#                 )
#             )

# -------------------------- QCHEM -------------------------- #

# from __future__ import annotations
# import cclib
# import os
# import mtr
# import subprocess
# from typing import Any, Iterable, Optional

# from ...workflow.tasks.task import Task

# __all__ = ["ExecuteVASP"]


# class ExecuteVASP(Task):
#     def __init__(
#         self,
#         output_name: str,
#         executable: str = "vasp_std",
#         work_directory: str = ".",
#         num_cores: int = 1,
#         parallel: bool = False,
#         handlers: Optional[Iterable[mtr.Handler]] = None,
#         name: Optional[str] = None,
#     ) -> None:
#         super().__init__()
#         self.settings["executable"] = executable
#         self.settings["output_path"] = os.path.join(
#             work_directory, mtr.expand(output_name)
#         )
#         self.settings["work_directory"] = mtr.expand(work_directory)

#         self.settings["num_cores"] = num_cores
#         self.settings["parallel"] = parallel

#         try:
#             os.makedirs(self.settings["work_directory"])
#         except FileExistsError:
#             pass

#     def run(self, **kwargs: Any) -> Any:
#         with open(self.settings["output_path"], "w") as f:
#             if self.settings["parallel"]:
#                 subprocess.call(
#                     [
#                         "mpirun",
#                         "np",
#                         str(self.settings["num_cores"]),
#                         self.settings["executable"],
#                     ],
#                     stdout=f,
#                     stderr=subprocess.STDOUT,
#                 )
#             else:
#                 subprocess.call([self.settings["executable"]], stdout=f)


class XTBOptimize(ExternalTask):
    def parse(self, output: str) -> mtr.Structure:
        with open(output, "r") as f:
            structure_file = re.search(
                r"optimized geometry written to:\s*(?P<structure>.*)\s*",
                "".join(f.readlines()),
            ).group("structure")

        return mtr.Structure.read(f"{pathlib.Path(output).parent}/{structure_file}")

    def run(self, structure: Union[mtr.Structure, str]) -> mtr.Structure:
        with self.io() as io:
            if isinstance(structure, mtr.Structure):
                with structure.tempfile(suffix=".xyz", dir=io.work_dir) as f:
                    self.engine.execute(f.name, self.io, arguments=["--opt"])
            else:
                self.engine.execute(structure, self.io, arguments=["--opt"])

            return self.parse(io.out)



from __future__ import annotations
import materia

__all__ = ["CCDCInput"]


class CCDCInput:
    """
    Interface to write inputs which can be run with CCDC.

    Attributes:
        ccdc_script (str): String representation of CCDC input.
    """

    def __init__(self, ccdc_script: str) -> None:
        self.ccdc_script = ccdc_script

    def __str__(self) -> str:
        """
        Get string representation of CCDC input.
        """
        return self.ccdc_script

    def write(self, filepath: str) -> None:
        """
        Write CCDC input to a file.

        Args:
            filepath: Path to file to which the input will be written. Can be an absolute or a relative path.
        """
        with open(materia.expand(filepath), "w") as f:
            f.write(str(self))



from __future__ import annotations
from typing import Iterable, TypeVar, Union

import inspect
import materia as mtr
import re

__all__ = ["MultiwfnOutput"]

T = TypeVar("T")


class MultiwfnOutput:
    def __init__(self, filepath: str) -> None:
        self.filepath = mtr.expand(filepath)

    def get(self, *quantity_names: str) -> Union[T, Iterable[T]]:
        method_dict = dict(inspect.getmembers(self, predicate=inspect.isroutine))

        with open(self.filepath, "r") as f:
            lines = "".join(f.readlines())

        quantities = tuple(
            method_dict[name.lower()](lines=lines) for name in quantity_names
        )

        if len(quantities) == 1:
            return quantities[0]
        return quantities

    def volume(self, lines: str) -> materia.Qty:
        # read off the volume in atomic units (i.e. cubic Bohr)
        volume_pattern = re.compile(
            r"\s*Molecular\s*volume\s*:\s*\d*\.\d*\s*Bohr\^3,\s*\(\s*(\d*\.\d*)\s*Angstrom\^3\s*,\s*\d*\.\d*\s*cm\^3\/mol\)\s*"
        )
        # Molecular volume:    0.000 Bohr^3, (    0.000 Angstrom^3,    0.000 cm^3/mol)
        # the last match is the molecular volume (Multiwfn likes to spit out
        # other volumes earlier in the script as well)
        *_, volume_str = volume_pattern.search(lines).groups()

        return float(volume_str) * mtr.angstrom ** 3



from __future__ import annotations
import materia
from typing import Iterable, Optional

__all__ = ["PackmolInput"]


class PackmolInput:
    def __init__(
        self,
        tolerance: float,
        filetype: str,
        output_name: str,
        instructions: Optional[Iterable[str]] = None,
    ) -> None:
        self.tolerance = tolerance
        self.filetype = filetype
        self.output_name = output_name
        self.instructions = instructions or []
        self.structures = []

    def add_structure(
        self,
        structure_filepath: str,
        number: int,
        instructions: Optional[Iterable[str]] = None,
    ) -> None:
        self.structures.append((structure_filepath, number, instructions or []))

    def __str__(self) -> str:
        return (
            f"tolerance {self.tolerance}\noutput {self.output_name}.{self.filetype}\nfiletype {self.filetype}\n\n"
            + "\n".join(
                f"structure {sfp}\n  number {n}\n"
                + "".join(f"  {i}\n" for i in inst)
                + "end structure\n"
                for sfp, n, inst in self.structures
            )
        )

    def write(self, filepath: str) -> None:
        with open(materia.expand(filepath), "w") as f:
            f.write(str(self))


# class PackmolInput:
#     def __init__(
#         self,
#         tolerance: float,
#         filetype: str,
#         output_name: str,
#         structures: Iterable[PackmolStructure],
#         instructions: Optional[Iterable[str]] = None,
#     ) -> None:
#         self.tolerance = tolerance
#         self.filetype = filetype
#         self.output_name = output_name
#         self.structures = structures
#         self.instructions = instructions or []

#     def _write_structure(self, structure, tmpdir):
#         while True:
#             filepath = os.path.join(tmpdir, f"{uuid.uuid4().hex}.{self.filetype}")
#             try:
#                 structure.write(materia.expand(filepath))  # , overwrite=False
#                 break
#             except FileExistsError:
#                 continue

#         return filepath

#     def _structure_blocks(self, tmpdir, structure, number, structure_instructions):
#         filepath = (
#             structure
#             if isinstance(structure, str)
#             else self._write_structure(structure=structure, tmpdir=tmpdir)
#         )

#         return (
#             f"structure {filepath}\n  number {number}\n"
#             + "".join(f"  {i}\n" for i in structure_instructions)
#             + "end structure\n"
#         )

#     def __str__(self) -> str:
#         with tempfile.TemporaryDirectory(".") as tmpdir:
#             # structure_blocks = tuple(
#             #     self._structure_blocks(
#             #         tmpdir=tmpdir, structure=s, number=n, structure_instructions=si
#             #     )
#             #     for s, n, si in zip(
#             #         self.structures, self.numbers, self.structure_instructions
#             #     )
#             # )
#             for s in self.structures:
#                 while s.filepath is None:
#                     filepath = os.path.join(
#                         tmpdir, f"{uuid.uuid4().hex}.{self.filetype}"
#                     )
#                     try:
#                         s.write(materia.expand(filepath))  # , overwrite=False
#                     except FileExistsError:
#                         continue

#             return (
#                 f"tolerance {self.tolerance}\noutput {self.output_name}.{self.filetype}\nfiletype {self.filetype}\n\n"
#                 + "\n".join(str(s) for s in self.structures)
#             )

#     def write(self, filepath: str) -> None:
#         with open(materia.expand(filepath), "w") as f:
#             f.write(str(self))


# class PackmolStructure:
#     def __init__(
#         self,
#         structure: Union[materia.Structure, str],
#         number: int,
#         instructions: Optional[Iterable[str]] = None,
#     ) -> None:
#         if isinstance(structure, str):
#             self.structure = None
#             self.filepath = materia.expand(structure)
#         else:
#             self.structure = structure
#             self.filepath = None
#         self.number = number
#         self.instructions = instructions or []

#     def __str__(self) -> str:
#         return (
#             f"structure {self.filepath}\n  number {self.number}\n"
#             + "".join(f"  {i}\n" for i in self.instructions)
#             + "end structure\n"
#         )

#     def write(self, filepath: str) -> None:
#         if self.structure is not None:
#             self.structure.write(materia.expand(filepath))
#             self.filepath = materia.expand(filepath)


# class PackmolInput:
#     def __init__(
#         self,
#         tolerance: float,
#         output_filename: str,
#         output_filetype=None,
#         structure_blocks=(),
#         restart_to=None,
#         restart_from=None,
#         discale=None,
#         maxit=None,
#         movebadrandom=False,
#         add_amber_ter=False,
#         add_box_sides=None,
#         sidemax=None,
#         seed=None,
#         randominitialpoint=False,
#         avoid_overlap=None,
#         nloop=None,
#         precision=None,
#         writeout=None,
#         writebad=False,
#         iprint1=None,
#         iprint2=None,
#         fbins=None,
#         chkgrad=False,
#         use_short_tol=False,
#         short_tol_dist=None,
#         short_tol_scale=None,
#     ) -> None:
#         """Short summary.
#
#         Parameters
#         ----------
#         tolerance : type
#             Required distance tolerance (for systems at room temperature and pressure and coordinates in Angstroms, 2.0 Å is a good value). Some users have observed that with the 2.0 Angs tolerance some cyclic systems (toluene, for example) may eventually get linked during packing. If you observe that kind of behaviour, a small increase on the distance tolerance probably will solve your problem. For coarse-grained models, which are not atomistic, the distance tolerance desired might be larger than that generally used for atomistic systems. However, this is not always the case, as the usual tolerance may well be large enough to guarantee a proper initial packing of the system.
#         output_filename : string
#             Name of the output file to be created.
#         output_filetype : string
#             Format of the output file (pdb, tinker, xyz, or moldy). If `output_filetype` is None, then it will be inferred from the extension of `output_filename`.
#         structure_blocks: type
#             StructureBlock objects whose contents should be added to the input file.
#         restart_to : type
#             This option tells Packmol to write restart information to the specified file.
#         restart_from : type
#             This option tells Packmol to read restart information from the specified file.
#         discale : type
#             This option controls the distance tolerance actually used in the local optimization method. It was found that using larger distances helps sometimes. Try setting discale to 1.5, for example.
#         maxit : type
#             This is the maximum number of iterations of the local optimizer (GENCAN) per loop. The default value is currently 20, changing it may improve (or worse) the convergence.
#         movebadrandom : type
#             One of the convergence heuristics of Packmol consists in moving molecules that are badly placed. If this option is set, the molecules will be placed in new random position in the box. If not (default), the molecules are moved to positions nearby molecules that are well packed. Using this option can help when the restraints are complex, but will probably be bad if there are large structures, because the new random position might overlap with those..
#         add_amber_ter : type
#             Add the TER flag betwen every molecule (AMBER uses this).
#         add_box_sides : type
#             Add box side information to output PDB File (GROMACS uses this), where the `1.0` is an optional real number that will be added to the length of each side, if the actual sides of your simulation box will not be exactly the same as the maximum and minimum coordinates of the molecules of the system (usually, if you are using periodic boundary conditions, you may want to add `1.0` to avoid clashes at the boundary).
#         sidemax : type
#             Increase maximum system dimensions. `sidemax` is used to build an initial approximation of the molecular distribution. Increase `sidemax` if your system is very large, or your system may look cut out by this maximum dimension. All system coordinates must fit within `-sidemax` and `+sidemax`, and using a `sidemax` that approximately fits your system may accelerate a little bit the packing calculation.
#         seed : type
#             Change random number generator seed. Use `-1` to generate a seed automatically from the computer time.
#         randominitialpoint : type
#             Use a truly random initial point for the minimization (the default option is to generate a roughly homogeneous-density initial approximation).
#         avoid_overlap : type
#             Avoid, or not, overlap with fixed molecules at initial point (avoiding this overlaps is generally recommended, but sometimes generates gaps that are too large).
#         nloop : type
#             Change the maximum number of optimization loops.
#         precision : type
#             Change the precision required for the solution: how close the solution must be to the desired distances to be considered correct.
#         writeout : type
#             Change the frequency of output file writing.
#         writebad : type
#             Write the current point to output file even if it is worse than the the best point so far (used for checking purposes only).
#         iprint1 : type
#             Change the optimization subroutine printing output.
#         iprint2 : type
#             Change the optimization subroutine printing output.
#         fbins : type
#             Change the number of bins of the linked-cell method (technical).
#         chkgrad : boolean
#             Compare analytical and finite-difference gradients: This is only for testing purposes. Writes chkgrad.log file containing the comparison.
#         use_short_tol : boolean
#             These options define a new penalty tolerance for short distances, to increase the repulsion between atoms when they are too close to each other. The "use" parameter turns on this feature, using default parameters (dist=tolerance/2; scale=3).
#         short_tol_dist : float
#              These options define a new penalty tolerance for short distances, to increase the repulsion between atoms when they are too close to each other. The "dist" parameter defines from which distance the penalty is applied. This affects the penalty of all atoms. Remember that the "tolerance" is twice the size of the "radius" of each atom.
#         short_tol_scale : float
#             These options define a new penalty tolerance for short distances, to increase the repulsion between atoms when they are too close to each other. The "scale" parameter how greater this penalty is relative to the default penalty. This affects the penalty of all atoms. Remember that the "tolerance" is twice the size of the "radius" of each atom.
#         """
#         self.tolerance = tolerance
#         self.output_filename = output_filename
#         self.output_filetype = output_filetype
#         self.structure_blocks = structure_blocks
#         self.restart_to = restart_to
#         self.restart_from = restart_from
#         self.discale = discale
#         self.maxit = maxit
#         self.movebadrandom = movebadrandom
#         self.add_amber_ter = add_amber_ter
#         self.add_box_sides = add_box_sides
#         self.sidemax = sidemax
#         self.seed = seed
#         self.randominitialpoint = randominitialpoint
#         self.avoid_overlap = avoid_overlap
#         self.nloop = nloop
#         self.precision = precision
#         self.writeout = writeout
#         self.writebad = writebad
#         self.iprint1 = iprint1
#         self.iprint2 = iprint2
#         self.fbins = fbins
#         self.chkgrad = chkgrad
#         self.use_short_tol = use_short_tol
#         self.short_tol_dist = short_tol_dist
#         self.short_tol_scale = short_tol_scale
#
#     def write(self, filepath: str) -> None:
#         with open(materia.expand(filepath), "w") as f:
#             f.write(str(self))
#
#     def __str__(self) -> str:
#         s = f"tolerance {self.tolerance}\n"
#         s += f"output {self.output_filename}\n"
#         if self.output_filetype is None:
#             *_, output_filetype = self.output_filename.split(".")
#             s += f"filetype {output_filetype}\n"
#         else:
#             s += f"filetype {self.output_filetype}\n"
#         s += f"restart_to {self.restart_to}\n" if self.restart_to is not None else ""
#         s += (
#             f"restart_from {self.restart_from}\n"
#             if self.restart_from is not None
#             else ""
#         )
#         s += f"discale {self.discale}\n" if self.discale is not None else ""
#         s += f"maxit {self.maxit}\n" if self.maxit is not None else ""
#         s += f"movebadrandom\n" if self.movebadrandom else ""
#         s += f"add_amber_ter\n" if self.add_amber_ter else ""
#         s += (
#             f"add_box_sides {self.add_box_sides}\n"
#             if self.add_box_sides is not None
#             else ""
#         )
#         s += (
#             f'sidemax {".d".join(self.sidemax.split("."))}\n'
#             if self.sidemax is not None
#             else ""
#         )
#         s += f"seed {self.seed}\n" if self.seed is not None else ""
#         s += f"randominitialpoint\n" if self.randominitialpoint else ""
#         s += (
#             f'avoid_overlap {"yes" if self.avoid_overlap else "no"}\n'
#             if self.avoid_overlap
#             else ""
#         )
#         s += f"nloop {self.nloop}\n" if self.nloop is not None else ""
#         s += f"precision {self.precision}\n" if self.precision is not None else ""
#         s += f"writeout {self.writeout}\n" if self.writeout is not None else ""
#         s += f"writebad\n" if self.writebad else ""
#         s += f"iprint1 {self.iprint1}\n" if self.iprint1 is not None else ""
#         s += f"iprint2 {self.iprint2}\n" if self.iprint2 is not None else ""
#         s += f"fbins {self.fbins}\n" if self.fbins is not None else ""
#         s += f"chkgrad {self.chkgrad}\n" if self.chkgrad else ""
#         s += f"use_short_tol {self.use_short_tol}\n" if self.use_short_tol else ""
#         s += (
#             f"short_tol_dist {self.short_tol_dist}\n"
#             if self.short_tol_dist is not None
#             else ""
#         )
#         s += (
#             f"short_tol_scale {self.short_tol_scale}\n"
#             if self.short_tol_scale is not None
#             else ""
#         )
#         s += "\n".join(str(sb) for sb in self.structure_blocks)
#         # FIXME: include all the other self attribute strings
#         return s
#
#
# class PackmolStructureBlock:
#     def __init__(
#         self,
#         molecule_structure_file,
#         number,
#         constraints=(),
#         atoms_blocks=(),
#         radius=None,
#         residue_number=None,
#         restart_to=None,
#         restart_from=None,
#         fscale=None,
#         short_radius=None,
#         short_radius_scale=None,
#         nloop=None,
#     ):
#         self.molecule_structure_file = materia.expand(molecule_structure_file)
#         self.number = number
#         self.constraints = constraints
#         self.atoms_blocks = atoms_blocks
#         self.radius = radius
#         self.residue_number = residue_number
#         self.restart_to = restart_to
#         self.restart_from = restart_from
#         self.fscale = fscale
#         self.short_radius = short_radius
#         self.short_radius_scale = short_radius_scale
#         self.nloop = nloop
#
#     def __str__(self) -> str:
#         s = "structure "
#         s += f"{self.molecule_structure_file}\n"
#         s += f"  number {self.number}\n"
#         s += "".join(f"  {str(c)}\n" for c in self.constraints)
#         s += "".join(f"  {str(ab)}\n" for ab in self.atoms_blocks)
#         s += f"  radius {radius}\n" if self.radius is not None else ""
#         s += (
#             f"  resnumber {self.residue_number}\n"
#             if self.residue_number is not None
#             else ""
#         )
#         s += f"  restart_to {self.restart_to}\n" if self.restart_to is not None else ""
#         s += (
#             f"  restart_to {self.restart_from}\n"
#             if self.restart_from is not None
#             else ""
#         )
#         s += f"  fscale {self.fscale}\n" if self.fscale is not None else ""
#         s += (
#             f"  short_radius {self.short_radius}\n"
#             if self.short_radius is not None
#             else ""
#         )
#         s += (
#             f"  short_radius_scale {self.short_radius_scale}\n"
#             if self.short_radius_scale is not None
#             else ""
#         )
#         s += f"  nloop {self.nloop}\n" if self.nloop is not None else ""
#         s += "end structure\n"
#
#         return s
#
#
# class PackmolAtomsBlock:
#     def __init(
#         self,
#         atoms,
#         constraints=(),
#         radius=None,
#         fscale=None,
#         short_radius=None,
#         short_radius_scale=None,
#     ):
#         self.atoms = atoms
#         self.constraints = constraints
#         self.radius = radius
#         self.fscale = fscale
#         self.short_radius = short_radius
#         self.short_radius_scale = short_radius_scale
#
#     def __str__(self) -> str:
#         atoms_str = " ".join(str(a) for a in atoms)
#         constraints_str = "".join(f"  {str(c)}\n" for c in self.constraints)
#         radius_str = f"  radius {radius}\n" if self.radius is not None else ""
#         fscale_str = f"  fscale {self.fscale}\n" if self.fscale is not None else ""
#         short_radius_str = (
#             f"  short_radius {self.short_radius}\n"
#             if self.short_radius is not None
#             else ""
#         )
#         short_radius_scale_str = (
#             f"  short_radius_scale {self.short_radius_scale}\n"
#             if self.short_radius_scale is not None
#             else ""
#         )
#
#         return (
#             f"atoms {atoms_str}\n"
#             + constraints_str
#             + radius_str
#             + fscale_str
#             + short_radius_str
#             + short_radius_scale_str
#             + "end atoms\n"
#         )
#
#
# # CONSTRAINTS
#
#
# class PackmolFixed:
#     def __init__(self, x, y, z, a, b, c) -> None:
#         self.x = x
#         self.y = y
#         self.z = z
#         self.a = a
#         self.b = b
#         self.c = c
#
#     def __str__(self) -> str:
#         return f"fixed {self.x} {self.y} {self.z} {self.a} {self.b} {self.c}"
#
#
# class PackmolCube:
#     def __init__(self, xmin, ymin, zmin, d, inside) -> None:
#         self.xmin = xmin
#         self.ymin = ymin
#         self.zmin = zmin
#         self.d = d
#         self.inside_outside = "inside" if inside else "outside"
#
#     def __str__(self) -> str:
#         return (
#             f"{self.inside_outside} cube {self.xmin} {self.ymin} {self.zmin} {self.d}"
#         )
#
#
# class PackmolBox:
#     def __init__(self, xmin, ymin, zmin, xmax, ymax, zmax, inside) -> None:
#         self.xmin = xmin
#         self.ymin = ymin
#         self.zmin = zmin
#         self.xmax = xmax
#         self.ymax = ymax
#         self.zmax = zmax
#         self.inside_outside = "inside" if inside else "outside"
#
#     def __str__(self) -> str:
#         return f"{self.inside_outside} box {self.xmin} {self.ymin} {self.zmin} {self.xmax} {self.ymax} {self.zmax}"
#
#
# class PackmolSphere:
#     def __init__(self, a, b, c, d, inside) -> None:
#         self.a = a
#         self.b = b
#         self.c = c
#         self.d = d
#         self.inside_outside = "inside" if inside else "outside"
#
#     def __str__(self) -> str:
#         return f"{self.inside_outside} sphere {self.a} {self.b} {self.c} {self.d}"
#
#
# class PackmolEllipsoid:
#     def __init__(self, a1, b1, c1, a2, b2, c2, d, inside) -> None:
#         self.a1 = a1
#         self.b1 = b1
#         self.c1 = c1
#         self.a2 = a2
#         self.b2 = b2
#         self.c2 = c2
#         self.d = d
#         self.inside_outside = "inside" if inside else "outside"
#
#     def __str__(self) -> str:
#         return f"{self.inside_outside} ellipsoid {self.a1} {self.b1} {self.c1} {self.a2} {self.b2} {self.c2} {self.d}"
#
#
# class PackmolPlane:
#     def __init__(self, a, b, c, d, over) -> None:
#         self.a = a
#         self.b = b
#         self.c = c
#         self.d = d
#         self.over_below = "over" if over else "below"
#
#     def __str__(self) -> str:
#         return f"{self.over_below} plane {self.a} {self.b} {self.c} {self.d}"
#
#
# class PackmolCylinder:
#     def __init__(self, a1, b1, c1, a2, b2, c2, d, l, inside) -> None:
#         self.a1 = a1
#         self.b1 = b1
#         self.c1 = c1
#         self.a2 = a2
#         self.b2 = b2
#         self.c2 = c2
#         self.d = d
#         self.l = l
#         self.inside_outside = "inside" if inside else "outside"
#
#     def __str__(self) -> str:
#         return f"{self.inside_outside} cylinder {self.a1} {self.b1} {self.c1} {self.a2} {self.b2} {self.c2} {self.d} {self.l}"
#
#
# class PackmolConstrainRotation:
#     def __init__(self, axis, angle, variance) -> None:
#         self.axis = axis
#         self.angle = angle
#         self.variance = variance
#
#     def __str__(self) -> str:
#         return f"constrain_rotation {self.axis} {self.angle} {self.variance}"
#
#
# # import materia
# #
# # class PackmolInput:
# #     def __init__(self, tolerance, output_filename, output_filetype=None, structure_blocks=(), restart_to=None, restart_from=None, discale=None,
# #                  maxit=None, movebadrandom=False, add_amber_ter=False, add_box_sides=None, sidemax=None, seed=None, randominitialpoint=False,
# #                  avoid_overlap=None, nloop=None, precision=None, writeout=None, writebad=False, iprint1=None, iprint2=None, fbins=None,
# #                  chkgrad=False, use_short_tol=False, short_tol_dist=None, short_tol_scale=None):
# #         """Short summary.
# #
# #         Parameters
# #         ----------
# #         tolerance : type
# #             Required distance tolerance (for systems at room temperature and pressure and coordinates in Angstroms, 2.0 Å is a good value). Some users have observed that with the 2.0 Angs tolerance some cyclic systems (toluene, for example) may eventually get linked during packing. If you observe that kind of behaviour, a small increase on the distance tolerance probably will solve your problem. For coarse-grained models, which are not atomistic, the distance tolerance desired might be larger than that generally used for atomistic systems. However, this is not always the case, as the usual tolerance may well be large enough to guarantee a proper initial packing of the system.
# #         output_filename : string
# #             Name of the output file to be created.
# #         output_filetype : string
# #             Format of the output file (pdb, tinker, xyz, or moldy). If `output_filetype` is None, then it will be inferred from the extension of `output_filename`.
# #         structure_blocks: type
# #             StructureBlock objects whose contents should be added to the input file.
# #         restart_to : type
# #             This option tells Packmol to write restart information to the specified file.
# #         restart_from : type
# #             This option tells Packmol to read restart information from the specified file.
# #         discale : type
# #             This option controls the distance tolerance actually used in the local optimization method. It was found that using larger distances helps sometimes. Try setting discale to 1.5, for example.
# #         maxit : type
# #             This is the maximum number of iterations of the local optimizer (GENCAN) per loop. The default value is currently 20, changing it may improve (or worse) the convergence.
# #         movebadrandom : type
# #             One of the convergence heuristics of Packmol consists in moving molecules that are badly placed. If this option is set, the molecules will be placed in new random position in the box. If not (default), the molecules are moved to positions nearby molecules that are well packed. Using this option can help when the restraints are complex, but will probably be bad if there are large structures, because the new random position might overlap with those..
# #         add_amber_ter : type
# #             Add the TER flag betwen every molecule (AMBER uses this).
# #         add_box_sides : type
# #             Add box side information to output PDB File (GROMACS uses this), where the `1.0` is an optional real number that will be added to the length of each side, if the actual sides of your simulation box will not be exactly the same as the maximum and minimum coordinates of the molecules of the system (usually, if you are using periodic boundary conditions, you may want to add `1.0` to avoid clashes at the boundary).
# #         sidemax : type
# #             Increase maximum system dimensions. `sidemax` is used to build an initial approximation of the molecular distribution. Increase `sidemax` if your system is very large, or your system may look cut out by this maximum dimension. All system coordinates must fit within `-sidemax` and `+sidemax`, and using a `sidemax` that approximately fits your system may accelerate a little bit the packing calculation.
# #         seed : type
# #             Change random number generator seed. Use `-1` to generate a seed automatically from the computer time.
# #         randominitialpoint : type
# #             Use a truly random initial point for the minimization (the default option is to generate a roughly homogeneous-density initial approximation).
# #         avoid_overlap : type
# #             Avoid, or not, overlap with fixed molecules at initial point (avoiding this overlaps is generally recommended, but sometimes generates gaps that are too large).
# #         nloop : type
# #             Change the maximum number of optimization loops.
# #         precision : type
# #             Change the precision required for the solution: how close the solution must be to the desired distances to be considered correct.
# #         writeout : type
# #             Change the frequency of output file writing.
# #         writebad : type
# #             Write the current point to output file even if it is worse than the the best point so far (used for checking purposes only).
# #         iprint1 : type
# #             Change the optimization subroutine printing output.
# #         iprint2 : type
# #             Change the optimization subroutine printing output.
# #         fbins : type
# #             Change the number of bins of the linked-cell method (technical).
# #         chkgrad : boolean
# #             Compare analytical and finite-difference gradients: This is only for testing purposes. Writes chkgrad.log file containing the comparison.
# #         use_short_tol : boolean
# #             These options define a new penalty tolerance for short distances, to increase the repulsion between atoms when they are too close to each other. The "use" parameter turns on this feature, using default parameters (dist=tolerance/2; scale=3).
# #         short_tol_dist : float
# #              These options define a new penalty tolerance for short distances, to increase the repulsion between atoms when they are too close to each other. The "dist" parameter defines from which distance the penalty is applied. This affects the penalty of all atoms. Remember that the "tolerance" is twice the size of the "radius" of each atom.
# #         short_tol_scale : float
# #             These options define a new penalty tolerance for short distances, to increase the repulsion between atoms when they are too close to each other. The "scale" parameter how greater this penalty is relative to the default penalty. This affects the penalty of all atoms. Remember that the "tolerance" is twice the size of the "radius" of each atom.
# #         """
# #         self.tolerance = tolerance
# #         self.output_filename = output_filename
# #         self.output_filetype = output_filetype
# #         self.structure_blocks = structure_blocks
# #         self.restart_to = restart_to
# #         self.restart_from = restart_from
# #         self.discale = discale
# #         self.maxit = maxit
# #         self.movebadrandom = movebadrandom
# #         self.add_amber_ter = add_amber_ter
# #         self.add_box_sides = add_box_sides
# #         self.sidemax = sidemax
# #         self.seed = seed
# #         self.randominitialpoint = randominitialpoint
# #         self.avoid_overlap = avoid_overlap
# #         self.nloop = nloop
# #         self.precision = precision
# #         self.writeout = writeout
# #         self.writebad = writebad
# #         self.iprint1 = iprint1
# #         self.iprint2 = iprint2
# #         self.fbins = fbins
# #         self.chkgrad = chkgrad
# #         self.use_short_tol = use_short_tol
# #         self.short_tol_dist = short_tol_dist
# #         self.short_tol_scale = short_tol_scale
# #
# #     def write(self, filepath):
# #         with open(materia.expand(filepath),'w') as f:
# #             f.write(str(self))
# #
# #     def __str__(self) -> str:
# #         s = f'tolerance {self.tolerance}\n'
# #         s += f'output {self.output_filename}\n'
# #         if self.output_filetype is None:
# #             *_,output_filetype = self.output_filename.split('.')
# #             s += f'filetype {output_filetype}\n'
# #         else:
# #             s += f'filetype {self.output_filetype}\n'
# #         s += f'restart_to {self.restart_to}\n' if self.restart_to is not None else ''
# #         s += f'restart_from {self.restart_from}\n' if self.restart_from is not None else ''
# #         s += f'discale {self.discale}\n' if self.discale is not None else ''
# #         s += f'maxit {self.maxit}\n' if self.maxit is not None else ''
# #         s += f'movebadrandom\n' if self.movebadrandom else ''
# #         s += f'add_amber_ter\n' if self.add_amber_ter else ''
# #         s += f'add_box_sides {self.add_box_sides}\n' if self.add_box_sides is not None else ''
# #         s += f'sidemax {".d".join(self.sidemax.split("."))}\n' if self.sidemax is not None else ''
# #         s += f'seed {self.seed}\n' if self.seed is not None else ''
# #         s += f'randominitialpoint\n' if self.randominitialpoint else ''
# #         s += f'avoid_overlap {"yes" if self.avoid_overlap else "no"}\n' if self.avoid_overlap else ''
# #         s += f'nloop {self.nloop}\n' if self.nloop is not None else ''
# #         s += f'precision {self.precision}\n' if self.precision is not None else ''
# #         s += f'writeout {self.writeout}\n' if self.writeout is not None else ''
# #         s += f'writebad\n' if self.writebad else ''
# #         s += f'iprint1 {self.iprint1}\n' if self.iprint1 is not None else ''
# #         s += f'iprint2 {self.iprint2}\n' if self.iprint2 is not None else ''
# #         s += f'fbins {self.fbins}\n' if self.fbins is not None else ''
# #         s += f'chkgrad {self.chkgrad}\n' if self.chkgrad else ''
# #         s += f'use_short_tol {self.use_short_tol}\n' if self.use_short_tol else ''
# #         s += f'short_tol_dist {self.short_tol_dist}\n' if self.short_tol_dist is not None else ''
# #         s += f'short_tol_scale {self.short_tol_scale}\n' if self.short_tol_scale is not None else ''
# #         s += '\n'.join(str(sb) for sb in self.structure_blocks)
# #         # FIXME: include all the other self attribute strings
# #         return s
# #
# # class PackmolStructureBlock:
# #     def __init__(self, molecule_structure_file, number, constraints=(), atoms_blocks=(), radius=None, residue_number=None, restart_to=None, restart_from=None,
# #                  fscale=None, short_radius=None, short_radius_scale=None, nloop=None):
# #         self.molecule_structure_file = materia.expand(molecule_structure_file)
# #         self.number = number
# #         self.constraints = constraints
# #         self.atoms_blocks = atoms_blocks
# #         self.radius = radius
# #         self.residue_number = residue_number
# #         self.restart_to = restart_to
# #         self.restart_from = restart_from
# #         self.fscale = fscale
# #         self.short_radius = short_radius
# #         self.short_radius_scale = short_radius_scale
# #         self.nloop = nloop
# #
# #     def __str__(self) -> str:
# #         s = 'structure '
# #         s += f'{self.molecule_structure_file}\n'
# #         s += f'  number {self.number}\n'
# #         s += ''.join(f'  {str(c)}\n' for c in self.constraints)
# #         s += ''.join(f'  {str(ab)}\n' for ab in self.atoms_blocks)
# #         s += f'  radius {radius}\n' if self.radius is not None else ''
# #         s += f'  resnumber {self.residue_number}\n' if self.residue_number is not None else ''
# #         s += f'  restart_to {self.restart_to}\n' if self.restart_to is not None else ''
# #         s += f'  restart_to {self.restart_from}\n' if self.restart_from is not None else ''
# #         s += f'  fscale {self.fscale}\n' if self.fscale is not None else ''
# #         s += f'  short_radius {self.short_radius}\n' if self.short_radius is not None else ''
# #         s += f'  short_radius_scale {self.short_radius_scale}\n' if self.short_radius_scale is not None else ''
# #         s += f'  nloop {self.nloop}\n' if self.nloop is not None else ''
# #         s += 'end structure\n'
# #
# #         return s
# #
# # class PackmolAtomsBlock:
# #     def __init(self, atoms, constraints=(), radius=None, fscale=None, short_radius=None, short_radius_scale=None):
# #         self.atoms = atoms
# #         self.constraints = constraints
# #         self.radius = radius
# #         self.fscale = fscale
# #         self.short_radius = short_radius
# #         self.short_radius_scale = short_radius_scale
# #
# #     def __str__(self) -> str:
# #         atoms_str = ' '.join(str(a) for a in atoms)
# #         constraints_str = ''.join(f'  {str(c)}\n' for c in self.constraints)
# #         radius_str = f'  radius {radius}\n' if self.radius is not None else ''
# #         fscale_str = f'  fscale {self.fscale}\n' if self.fscale is not None else ''
# #         short_radius_str = f'  short_radius {self.short_radius}\n' if self.short_radius is not None else ''
# #         short_radius_scale_str = f'  short_radius_scale {self.short_radius_scale}\n' if self.short_radius_scale is not None else ''
# #
# #         return f'atoms {atoms_str}\n' + constraints_str + radius_str + fscale_str + short_radius_str + short_radius_scale_str + 'end atoms\n'
# #
# # # CONSTRAINTS
# #
# # class PackmolFixed:
# #     def __init__(self, x, y, z, a, b, c) -> None:
# #         self.x = x
# #         self.y = y
# #         self.z = z
# #         self.a = a
# #         self.b = b
# #         self.c = c
# #
# #     def __str__(self) -> str:
# #         return f'fixed {self.x} {self.y} {self.z} {self.a} {self.b} {self.c}'
# #
# # class PackmolCube:
# #     def __init__(self, xmin, ymin, zmin, d, inside) -> None:
# #         self.xmin = xmin
# #         self.ymin = ymin
# #         self.zmin = zmin
# #         self.d = d
# #         self.inside_outside = 'inside' if inside else 'outside'
# #
# #     def __str__(self) -> str:
# #         return f'{self.inside_outside} cube {self.xmin} {self.ymin} {self.zmin} {self.d}'
# #
# # class PackmolBox:
# #     def __init__(self, xmin, ymin, zmin, xmax, ymax, zmax, inside) -> None:
# #         self.xmin = xmin
# #         self.ymin = ymin
# #         self.zmin = zmin
# #         self.xmax = xmax
# #         self.ymax = ymax
# #         self.zmax = zmax
# #         self.inside_outside = 'inside' if inside else 'outside'
# #
# #     def __str__(self) -> str:
# #         return f'{self.inside_outside} box {self.xmin} {self.ymin} {self.zmin} {self.xmax} {self.ymax} {self.zmax}'
# #
# # class PackmolSphere:
# #     def __init__(self, a, b, c, d, inside) -> None:
# #         self.a = a
# #         self.b = b
# #         self.c = c
# #         self.d = d
# #         self.inside_outside = 'inside' if inside else 'outside'
# #
# #     def __str__(self) -> str:
# #         return f'{self.inside_outside} sphere {self.a} {self.b} {self.c} {self.d}'
# #
# # class PackmolEllipsoid:
# #     def __init__(self, a1, b1, c1, a2, b2, c2, d, inside) -> None:
# #         self.a1 = a1
# #         self.b1 = b1
# #         self.c1 = c1
# #         self.a2 = a2
# #         self.b2 = b2
# #         self.c2 = c2
# #         self.d = d
# #         self.inside_outside = 'inside' if inside else 'outside'
# #
# #     def __str__(self) -> str:
# #         return f'{self.inside_outside} ellipsoid {self.a1} {self.b1} {self.c1} {self.a2} {self.b2} {self.c2} {self.d}'
# #
# # class PackmolPlane:
# #     def __init__(self, a, b, c, d, over) -> None:
# #         self.a = a
# #         self.b = b
# #         self.c = c
# #         self.d = d
# #         self.over_below = 'over' if over else 'below'
# #
# #     def __str__(self) -> str:
# #         return f'{self.over_below} plane {self.a} {self.b} {self.c} {self.d}'
# #
# # class PackmolCylinder:
# #     def __init__(self, a1, b1, c1, a2, b2, c2, d, l, inside) -> None:
# #         self.a1 = a1
# #         self.b1 = b1
# #         self.c1 = c1
# #         self.a2 = a2
# #         self.b2 = b2
# #         self.c2 = c2
# #         self.d = d
# #         self.l = l
# #         self.inside_outside = 'inside' if inside else 'outside'
# #
# #     def __str__(self) -> str:
# #         return f'{self.inside_outside} cylinder {self.a1} {self.b1} {self.c1} {self.a2} {self.b2} {self.c2} {self.d} {self.l}'
# #
# # class PackmolConstrainRotation:
# #     def __init__(self, axis, angle, variance) -> None:
# #         self.axis = axis
# #         self.angle = angle
# #         self.variance = variance
# #
# #     def __str__(self) -> str:
# #         return f'constrain_rotation {self.axis} {self.angle} {self.variance}'
